      subroutine rbea(n,m,a,istart,b,ib,ifin)
c     Reorder rows using BEA, bond energy algorithm.
      dimension a(n,m), b(n,m), ib(n), ifin(n)
c------------------------------------------------------------------------------
c     a(n,m)         input matrix, rows of which are to be permuted
c     istart         1st row to be placed
c     b(n,m)         permuted rows to be stored in this array
c     ib(n)          integer list giving permutation carried out
c     ifin(n)        book-keeping vector: is row still active, or has it been 
c                    placed (resp. = 1, = 0).
c------------------------------------------------------------------------------
c     Bond energy algorithm -- see:
c
c     (1) W.T. McCormick, P.J. Schweitzer and T.W. White, 
c         "Problem decomposition and data reorganization by a clustering 
c         technique", Oper. Res., vol. 20, pp. 993-1009, Sept./Oct. 1972.
c     (2) P. Arabie and L.J. Hubert, 
c         "The bond energy algorithm revisited", IEEE Trans. Syst. Man 
c         Cybern., vol. 20, pp. 268-274, 1990.
c     (3) P. Arabie, S. Schleutermann, J. Daws and L. Hubert, 
c         "Marketing applications of sequencing and partitioning of 
c         nonsymmetric and/or two-mode matrices", in W. Gaul and M. Schader, 
c         Eds., Data Analysis, Decision Support, and Expert Knowledge 
c         Representation in Marketing, Springer Verlag, 1988, pp. 215-224.
c  
c     Implemented by F. Murtagh, Sept. 1991.
c------------------------------------------------------------------------------
c
c     Flags to indicate if row already chosen; 1 = not yet chosen/placed.
      do 200 i = 1, n
 200     ifin(i) = 1
c
c     Anticipate 1st placement. 'nplace' = # rows placed. 'nrem' = # remaining.
      nplace = 1
      nrem   = n-1
c
c     Place 1st row
      do 300 j = 1, m
         b(1,j)     = a(istart,j)
         ifin(istart) = 0
         ib(nplace) = istart
 300  continue
c
 400  continue
      sim   = -100000.0
      insrt = 0
c     'nplace' rows have been placed.
c     Now want next placement.  Have 'nrem' rows still to place.
      do 900 irow = 1, n
         if (ifin(irow).eq.1) then
c           For all still-to-be-placed rows...
c           1. Place right at beg.:
               sim1 = 0.0
               do 500 j = 1, m
                  sim1 = sim1 + a(irow,j)*b(1,j)
  500          continue
c           2. Place right at end:
               if (nplace.gt.1) then
                  sim2 = 0.0
                  do 600 j = 1, m
                     sim2 = sim2 + a(irow,j)*b(nplace,j)
  600             continue
               endif
c           3. Place between k and k+1, where k = 1, ..., nplace-1:
               if (nplace.ge.2) then
                  do 800 k = 1, nplace-1
c                    Path length involves sim with k'th and with k+1'th rows
c                    in 'b'; i.e. b(k,j) and b(k+1,j), for all j.
c                    Sim is with a(irow,j), for all j.
                     sim3 = 0.0
                     do 700 j = 1, m
                        sim3 = sim3 + a(irow,j)*(b(k,j)+b(k+1,j))
  700                continue
                     if (sim3.gt.sim) then
                        sim = sim3
                        insrt = k
                        iplrow = irow
                     endif
  800             continue
               endif
c
c              Scale up 'sim1' and 'sim2' relative to 'sim', since former
c              are based on one link only
               sim1 = sim1*2.0
               sim2 = sim2*2.0
c              Use 'sim' and 'insrt' to store final info on row to place.
               if (sim1.gt.sim) then
                  sim = sim1
                  insrt = 0
                  iplrow = irow
               endif
c              .ge. in following, to force tied case to end
               if (sim2.ge.sim) then
                  sim = sim2
                  insrt = nplace+1
                  iplrow = irow
               endif
            endif
 900     continue
c
c      So now, we want to make placement in location 'insrt+1'
c      1. This happens to be right at beginning:
       if (insrt.eq.0) then
c         Shift right
          do 1100 l = nplace+1, 2, -1
             ib(l)  = ib(l-1)
             do 1000 j = 1, m
                b(l,j) = b(l-1,j)
 1000        continue
 1100     continue
          do 1200 j = 1, m
             b(1,j) = a(iplrow,j)
 1200     continue
          ifin(iplrow) = 0
          nplace       = nplace + 1
          nrem         = nrem - 1
          ib(1)        = iplrow
          goto 1900
       endif
c
c      2. Placement happens to be right at end of all current placements:
       if (insrt.eq.nplace+1) then
c         Insert after all current placements.
          do 1300 j = 1, m
             b(nplace+1,j) = a(iplrow,j)
 1300     continue
          ifin(iplrow) = 0
          nplace       = nplace + 1
          nrem         = nrem -1
          ib(nplace)   = iplrow
          goto 1900
       endif
c
c      3. If we get to here, new placement is somewhere in the middle.
c      Shift rows 'insrt+1' to 'nplace', in 'b', right.
       do 1500 l = nplace+1, insrt+2, -1
          ib(l) = ib(l-1)
          do 1400 j = 1, m
             b(l,j) = b(l-1,j)
 1400     continue
 1500  continue
       do 1600 j = 1, m
          b(insrt+1,j) = a(iplrow,j)
 1600  continue
       nplace       = nplace + 1
       nrem         = nrem -1
       ifin(iplrow) = 0
       ib(insrt+1)   = iplrow
       goto 1900
c
 1900   continue
       if (nrem.ge.1) goto 400
c
       return
       end
                 
c------------------------------------------------------------------------------
      subroutine cbea(n,m,a,jstart,b,jb,jfin)
c     Reoder cols. using BEA, bond energy algorithm.
c     See references at beg. of routine 'rbea'.
      dimension a(n,m), b(n,m), jb(m), jfin(m)
c------------------------------------------------------------------------------
c     a(n,m)         input matrix, cols. of which are to be permuted
c     jstart         1st col. to be placed
c     b(n,m)         permuted cols. to be stored in this array
c     jb(m)          integer list giving permutation carried out
c     jfin(m)        book-keeping vector: is col. still active, or has it been 
c                    placed (resp. = 1, = 0).
c------------------------------------------------------------------------------
c
c     Flags to indicate if col. already chosen
      do 200 j = 1, m
 200     jfin(j) = 1
c
c     'nplace' cols. placed (anticipating!); 'nrem' cols. still to be placed.
      nplace = 1
      nrem   = m-1
c
c     Place 1st col.
      do 300 i = 1, n
         b(i,1)     = a(i,jstart)
         jfin(jstart) = 0
         jb(nplace) = jstart
 300  continue
c
 400  continue
      sim   = -100000.0
      insrt = 0
c     'nplace' cols. have been placed.
c     Now want next placement.  Have 'nrem' cols. still to place.
      do 900 jcol = 1, m
         if (jfin(jcol).eq.1) then
c           For all still-to-be-placed cols...
c           1. Place right at beg.:
               sim1 = 0.0
               do 500 i = 1, n
                  sim1 = sim1 + a(i,jcol)*b(i,1)
  500          continue
c           2. Place right at end:
               if (nplace.gt.1) then
                  sim2 = 0.0
                  do 600 i = 1, n
                     sim2 = sim2 + a(i,jcol)*b(i,nplace)
  600             continue
               endif
c           3. Place between k and k+1, where k = 1, ..., nplace-1:
               if (nplace.ge.2) then
                  do 800 k = 1, nplace-1
c                    Path length involves sim with k'th and with k+1'th cols.
c                    in 'b'; i.e. b(i,k) and b(i,k+1), for all i.
c                    Sim is with a(i,jcol), for all i.
                     sim3 = 0.0
                     do 700 i = 1, n
                        sim3 = sim3 + a(i,jcol)*(b(i,k)+b(i,k+1))
  700                continue
                     if (sim3.gt.sim) then
                        sim = sim3
                        insrt = k
                        jplcol = jcol
                     endif
  800             continue
               endif
c
c              Scale up 'sim1' and 'sim2' rel. to 'sim' since former are
c              based on only one link
               sim1 = 2.0*sim1
               sim2 = 2.0*sim2
c              Use 'sim' and 'insrt' to store final info. on col. to be placed.
               if (sim1.gt.sim) then
                  sim = sim1
                  insrt = 0
                  jplcol = jcol
               endif
c              .ge. in following, to force tied case to end
               if (sim2.ge.sim) then
                  sim = sim2
                  insrt = nplace+1
                  jplcol = jcol
               endif
            endif
 900     continue
c
c      So now, we want to make placement in location 'insrt+1'.
c      1. This happens to be right at beginning.
       if (insrt.eq.0) then
c         Shift right
          do 1100 l = nplace+1, 2, -1
             jb(l) = jb(l-1)
             do 1000 i = 1, n
                b(i,l) = b(i,l-1)
 1000        continue
 1100     continue
          do 1200 i = 1, n
             b(i,1) = a(i,jplcol)
 1200     continue
          jfin(jplcol) = 0
          nplace       = nplace + 1
          nrem         = nrem - 1
          jb(1)   = jplcol
          goto 1900
       endif
c
c      2. Placement happens to be right at end of already placed cols.
       if (insrt.eq.nplace+1) then
          do 1300 i = 1, n
             b(i,nplace+1) = a(i,jplcol)
 1300     continue
          jfin(jplcol) = 0
          nplace       = nplace + 1
          nrem         = nrem -1
          jb(nplace)   = jplcol
          goto 1900
       endif
c
c      3. New col. placement happens to be somewhere in the middle.
c      Shift cols. 'insrt+1' to 'nplace', in 'b', right.
       do 1500 l = nplace+1, insrt+2, -1
          jb(l)  = jb(l-1)
          do 1400 i = 1, n
             b(i,l) = b(i,l-1)
 1400     continue
 1500  continue
       do 1600 i = 1, n
          b(i,insrt+1) = a(i,jplcol)
 1600  continue
       nplace       = nplace + 1
       nrem         = nrem -1
       jfin(jplcol) = 0
       jb(insrt+1)   = jplcol
       goto 1900
c
 1900   continue
       if (nrem.ge.1) goto 400
c
       return
       end
c--------------------------------------------------------------------------
       subroutine energy(n,m,b,ener)
       dimension b(n,m)
c      Det. "bond energy" of array b
c      I.e. product of each elt. with its 4 nearest neighbors,
c      summed over all elts.
       ener = 0.0
c      Corner elts.
       ener = ener + b(1,1)*(b(1,2)+b(2,1))
       ener = ener + b(1,m)*(b(1,m-1)+b(2,m))                 
       ener = ener + b(n,1)*(b(n-1,1)+b(n,2))
       ener = ener + b(n,m)*(b(n-1,m)+b(n,m-1))
c      Next non-corner border elts.
       do 100 j = 2, m-1
          ener = ener + b(1,j)*(b(1,j-1)+b(1,j+1)+b(2,j))
          ener = ener + b(n,j)*(b(n,j-1)+b(n,j+1)+b(n-1,j))
 100   continue
       do 200 i = 2, n-1
          ener = ener + b(i,1)*(b(i-1,1)+b(i+1,1)+b(i,2))
          ener = ener + b(i,m)*(b(i-1,m)+b(i+1,m)+b(i,m-1))
 200   continue
c      Finally, all non-border elts.
       do 400 i = 2, n-1
          do 300 j = 2, m-1
             ener = ener + b(i,j)*(b(i-1,j)+b(i+1,j)+b(i,j-1)+b(i,j+1))
 300      continue
 400   continue
c
       return
       end
       
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C
C  CORRESPONDENCE ANALYSIS                                                   C
C                                                                            C
C  To call:   CALL CA(N,M,DATA,A1,W1,W2,A2,R,C,RCN,CCN,NF,IERR) where        C
C                                                                            C
C  N, M  : integer dimensions of ...                                         C
C  DATA  : input data.                                                       C
C          On output, DATA contains in first NF columns the projections of   C
C          the row-points on the first NF factors.                           C
C  A1    : matrix to be diagonalized, of dimensions M * M.                   C
C          On output, A1 contains in the first NF columns the projections of C
C          the column-points on the first NF factors.                        C
C  W1,W2 : real vectors of dimension M (see called routines for use).        C
C          On output, W1 contains the eigenvalues (in increasing order of    C
C          magnitude).                                                       C
C  A2    : real array of dimensions M * M (see called routines for use).     C
C  R, C  : vectors of length N and M (respectively) containing, on output,   C
C          the masses of the row and column points.                          C
C  RCN   : N * M matrix for row contributions (mass*projection**2).          C
C  CCN   : M * M matrix for col contributions.                               C
C  NF    : integer, no. of factors requested.                                C
C  IERR  : error indicator (normally zero).                                  C
C                                                                            C
C  Inputs here are N, M, DATA, METHOD, IPRINT (and IERR).                    C
C  Output information is contained in DATA, A1, W1, R and C.                 C
C  All printed outputs are carried out in easily recognizable subroutines    C
C  called from the first subroutine following.                               C
C                                                                            C
C  F. Murtagh, ST-ECF/ESA/ESO, Garching bei Muenchen, January 1986.          C
C                                                                            C 
C  HISTORY                                                                   C
C                                                                            C
C  Altered subr. names (prec. by "C") to avoid potential pbs. in linking     C
C  with subr.s of same names in other programs.             F.M., Aug. 1990  C
C  Added RCN and CCN (row/col. contributions) as returnable matrices;        C
C  formerly this information output only.                   F.M., Aug. 1990  C
C  Added NF as parameter, + assoc. changes in code.         F.M., Aug. 1992  C
C----------------------------------------------------------------------------C
        SUBROUTINE CA(N,M,DATA,A,W,FV1,Z,R,C,RCN,CCN,NF,IERR)
        REAL    DATA(N,M), A(M,M), W(M), FV1(M), Z(M,M), R(N), C(M)
        REAL    RCN(N,M), CCN(M,M)

C          Form row sums and total

        TOT = 0.0
        DO 200 I = 1, N
           R(I) = 0.0
           DO 100 J = 1, M
              TOT = TOT + DATA(I,J)
              R(I) = R(I) + DATA(I,J)
  100      CONTINUE
  200   CONTINUE

C       Form column sums and then means

        DO 400 J = 1, M
           C(J) = 0.0
           DO 300 I = 1, N
              C(J) = C(J) + DATA(I,J)
  300      CONTINUE
           IF (C(J).GT.0.0) GOTO 350
              IERR = 2
              RETURN
  350      CONTINUE
           C(J) = C(J)/TOT
  400   CONTINUE

C       Form row means and make data into frequencies

        DO 600 I = 1, N
           IF (R(I).GT.0.0) GOTO 450
              IERR = 2
              RETURN
  450      CONTINUE
           R(I) = R(I)/TOT
           DO 500 J = 1, M
              DATA(I,J) = DATA(I,J)/TOT
  500      CONTINUE
  600   CONTINUE

C       Form matrix to be diagonalized

        DO 900 J1 = 1, M
           DO 800 J2 = 1, M
              A(J1,J2) = 0.0
              DO 700 I = 1, N
                 A(J1,J2) = A(J1,J2) + DATA(I,J1)*DATA(I,J2)/
     X                                (R(I)*SQRT(C(J1)*C(J2)))
  700         CONTINUE
  800      CONTINUE
  900   CONTINUE

C       Carry out eigenreduction.

        M2 = M
        CALL CTRED2(M,M2,A,W,FV1,Z)
        CALL CTQL2(M,M2,W,FV1,Z,IERR)
        IF (IERR.NE.0) GOTO 9000

C       Determine factors from the eigenvectors of the symmetric
C       matrix which has been diagonalized.

        DO 1100 J1 = 1, M
           DO 1000 J2 = 1, M
              Z(J1,J2) = Z(J1,J2)/SQRT(C(J1))
 1000      CONTINUE
 1100   CONTINUE

C       Determine projections and contributions.
        CALL CPROJX(N,M,DATA,Z,FV1,R,NF)
        CALL CPROJY(M,W,A,Z,FV1,C,NF)

        CALL COUTCX(N,M,DATA,R,RCN,NF)
        CALL COUTCY(M,A,C,CCN,NF)

 9000   RETURN  
        END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C
C Reduce a real, symmetric matrix to a symmetric, tridiagonal                C
C matrix.                                                                    C 
C                                                                            C
C To call:    CALL CTRED2(NM,N,A,D,E,Z)    where                             C
C                                                                            C
C NM = row dimension of A and Z;                                             C
C N = order of matrix A (will always be <= NM);                              C
C A = symmetric matrix of order N to be reduced to tridiagonal form;         C
C D = vector of dim. N containing, on output, diagonal elts. of trid. matrix;C
C E = working vector of dim. at least N-1 to contain subdiagonal elts.;      C
C Z = matrix of dims. NM by N containing, on output, orthogonal              C
C                    transformation matrix producting the reduction.         C
C                                                                            C
C Normally a call to TQL2 will follow the call to TRED2 in order to          C
C produce all eigenvectors and eigenvalues of matrix A.                      C
C                                                                            C
C Algorithm used: Martin et al., Num. Math. 11, 181-195, 1968.               C
C                                                                            C
C Reference: Smith et al., Matrix Eigensystem Routines - EISPACK             C
C Guide, Lecture Notes in Computer Science 6, Springer-Verlag, 1976,         C
C pp. 489-494.                                                               C
C                                                                            C
C F. Murtagh, ST-ECF/ESA/ESO, Garching bei Muenchen, January 1986.           C
C----------------------------------------------------------------------------C
        SUBROUTINE CTRED2(NM,N,A,D,E,Z)
C
        REAL A(NM,N),D(N),E(N),Z(NM,N)
C
        DO 100 I = 1, N
           DO 100 J = 1, I
              Z(I,J) = A(I,J)
  100   CONTINUE
        IF (N.EQ.1) GOTO 320
        DO 300 II = 2, N
           I = N + 2 - II
           L = I - 1
           H = 0.0
           SCALE = 0.0
           IF (L.LT.2) GOTO 130
           DO 120 K = 1, L
              SCALE = SCALE + ABS(Z(I,K))
  120      CONTINUE
           IF (SCALE.NE.0.0) GOTO 140
  130      E(I) = Z(I,L)
           GOTO 290
  140      DO 150 K = 1, L
              Z(I,K) = Z(I,K)/SCALE
              H = H + Z(I,K)*Z(I,K)
  150      CONTINUE
C
           F = Z(I,L)
           G = -SIGN(SQRT(H),F)
           E(I) = SCALE * G
           H = H - F * G
           Z(I,L) = F - G
           F = 0.0
C
           DO 240 J = 1, L
              Z(J,I) = Z(I,J)/H
              G = 0.0
C             Form element of A*U.
              DO 180 K = 1, J
                 G = G + Z(J,K)*Z(I,K)
  180         CONTINUE
              JP1 = J + 1
              IF (L.LT.JP1) GOTO 220
              DO 200 K = JP1, L
                 G = G + Z(K,J)*Z(I,K)
  200         CONTINUE
C             Form element of P where P = I - U U' / H .
  220         E(J) = G/H
              F = F + E(J) * Z(I,J)
  240      CONTINUE
           HH = F/(H + H)
C          Form reduced A.
           DO 260 J = 1, L
              F = Z(I,J)
              G = E(J) - HH * F
              E(J) = G
              DO 250 K = 1, J
                 Z(J,K) = Z(J,K) - F*E(K) - G*Z(I,K)
  250         CONTINUE
  260      CONTINUE
  290      D(I) = H
  300   CONTINUE
  320   D(1) = 0.0
        E(1) = 0.0
C       Accumulation of transformation matrices.
        DO 500 I = 1, N
           L = I - 1
           IF (D(I).EQ.0.0) GOTO 380
           DO 360 J = 1, L
              G = 0.0
              DO 340 K = 1, L
                 G = G + Z(I,K) * Z(K,J)
  340         CONTINUE
              DO 350 K = 1, L
                 Z(K,J) = Z(K,J) - G * Z(K,I)
  350         CONTINUE
  360      CONTINUE
  380      D(I) = Z(I,I)
           Z(I,I) = 1.0
           IF (L.LT.1) GOTO 500
           DO 400 J = 1, L
              Z(I,J) = 0.0
              Z(J,I) = 0.0
  400      CONTINUE
  500   CONTINUE
C
        RETURN
        END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C
C Determine eigenvalues and eigenvectors of a symmetric,                     C
C tridiagonal matrix.                                                        C
C                                                                            C
C To call:    CALL CTQL2(NM,N,D,E,Z,IERR)    where                           C
C                                                                            C
C NM = row dimension of Z;                                                   C
C N = order of matrix Z;                                                     C
C D = vector of dim. N containing, on output, eigenvalues;                   C
C E = working vector of dim. at least N-1;                                   C
C Z = matrix of dims. NM by N containing, on output, eigenvectors;           C
C IERR = error, normally 0, but 1 if no convergence.                         C
C                                                                            C
C Normally the call to TQL2 will be preceded by a call to TRED2 in           C
C order to set up the tridiagonal matrix.                                    C
C                                                                            C
C Algorithm used: QL method of Bowdler et al., Num. Math. 11,                C
C 293-306, 1968.                                                             C
C                                                                            C
C Reference: Smith et al., Matrix Eigensystem Routines - EISPACK             C
C Guide, Lecture Notes in Computer Science 6, Springer-Verlag, 1976,         C
C pp. 468-474.                                                               C
C                                                                            C
C F. Murtagh, ST-ECFESA/ESO, Garching bei Muenchen, January 1986.            C
C----------------------------------------------------------------------------C
        SUBROUTINE CTQL2(NM,N,D,E,Z,IERR)
C
        REAL    D(N), E(N), Z(NM,N)
        DATA    EPS/1.E-12/
C
        IERR = 0
        IF (N.EQ.1) GOTO 1001
        DO 100 I = 2, N
           E(I-1) = E(I)
  100   CONTINUE
        F = 0.0
        B = 0.0
        E(N) = 0.0
C
        DO 240 L = 1, N
           J = 0
           H = EPS * (ABS(D(L)) + ABS(E(L)))
           IF (B.LT.H) B = H
C          Look for small sub-diagonal element.
           DO 110 M = L, N
              IF (ABS(E(M)).LE.B) GOTO 120
C             E(N) is always 0, so there is no exit through the bottom 
C             of the loop.
  110      CONTINUE
  120      IF (M.EQ.L) GOTO 220
  130      IF (J.EQ.30) GOTO 1000
           J = J + 1
C          Form shift.
           L1 = L + 1
           G = D(L)
           P = (D(L1)-G)/(2.0*E(L))
           R = SQRT(P*P+1.0)
           D(L) = E(L)/(P+SIGN(R,P))
           H = G-D(L)
C
           DO 140 I = L1, N
              D(I) = D(I) - H
  140      CONTINUE
C
           F = F + H
C          QL transformation.
           P = D(M)
           C = 1.0
           S = 0.0
           MML = M - L
C
           DO 200 II = 1, MML
              I = M - II
              G = C * E(I)
              H = C * P
              IF (ABS(P).LT.ABS(E(I))) GOTO 150
              C = E(I)/P
              R = SQRT(C*C+1.0)
              E(I+1) = S * P * R
              S = C/R
              C = 1.0/R
              GOTO 160
  150         C = P/E(I)
              R = SQRT(C*C+1.0)
              E(I+1) = S * E(I) * R
              S = 1.0/R
              C = C * S
  160         P = C * D(I) - S * G
              D(I+1) = H + S * (C * G + S * D(I))
C             Form vector.
              DO 180 K = 1, N
                 H = Z(K,I+1)
                 Z(K,I+1) = S * Z(K,I) + C * H
                 Z(K,I) = C * Z(K,I) - S * H
  180         CONTINUE
  200      CONTINUE
           E(L) = S * P
           D(L) = C * P
           IF (ABS(E(L)).GT.B) GOTO 130
  220      D(L) = D(L) + F
  240   CONTINUE
C
C       Order eigenvectors and eigenvalues.
        DO 300 II = 2, N
           I = II - 1
           K = I
           P = D(I)
           DO 260 J = II, N
              IF (D(J).GE.P) GOTO 260
              K = J
              P = D(J)
  260      CONTINUE
           IF (K.EQ.I) GOTO 300
           D(K) = D(I)
           D(I) = P
           DO 280 J = 1, N
              P = Z(J,I)
              Z(J,I) = Z(J,K)
              Z(J,K) = P
  280      CONTINUE
  300   CONTINUE
C
        GOTO 1001
C       Set error - no convergence to an eigenvalue after 30 iterations.
 1000   IERR = L
 1001   RETURN
        END  
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C
C  Output eigenvalues in order of decreasing value.                          C
C  The first eigenvalue in Correspondence Analysis is trivially              C
C  of value 1, and so is ignored.                                            C
C----------------------------------------------------------------------------C
        SUBROUTINE COUTEV(NVALS,VALS)
        DIMENSION       VALS(NVALS)

        TOT = 0.0
        DO 100 K = 1, NVALS-1
           TOT = TOT + VALS(K)
  100   CONTINUE

        CUM = 0.0
        K = NVALS 

  200   CONTINUE
        K = K - 1
        CUM = CUM + VALS(K)
        VPC = VALS(K) * 100.0 / TOT
        VCPC = CUM * 100.0 / TOT

        IF (K.GT.1) GOTO 200

        RETURN
        END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C
C   Form projections of row-points on first NF factors.                      C
C----------------------------------------------------------------------------C
        SUBROUTINE CPROJX(N,M,DATA,EVEC,VEC,R,NF)
        REAL    DATA(N,M), EVEC(M,M), VEC(M), R(N)

        NUM = NF
        DO 300 K = 1, N
           DO 50 L = 1, M
              VEC(L) = DATA(K,L)
   50      CONTINUE
           DO 200 I = 1, NUM
              DATA(K,I) = 0.0
              DO 100 J = 1, M
                 DATA(K,I) = DATA(K,I) + VEC(J) * EVEC(J,M-I)
  100         CONTINUE
              IF (R(K).GT.0.0) DATA(K,I) = DATA(K,I)/R(K)
              IF (R(K).EQ.0.0) DATA(K,I) = 0.0
  200      CONTINUE
  300   CONTINUE

        RETURN
        END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C
C   Form projections of column points on first NF factors.                   C
C----------------------------------------------------------------------------C
        SUBROUTINE CPROJY(M,EVALS,A,Z,VEC,C,NF)
        REAL    EVALS(M), A(M,M), Z(M,M), VEC(M), C(M)

        NUM = NF
        DO 300 J1 = 1, M
           DO 50 L = 1, M
              VEC(L) = A(J1,L)
   50      CONTINUE
           DO 200 J2 = 1, NUM
              A(J1,J2) = 0.0
              DO 100 J3 = 1, M
                 A(J1,J2) = A(J1,J2) + VEC(J3)*Z(J3,M-J2)*
     X                                   SQRT(C(J3))
  100         CONTINUE
              IF ((EVALS(M-J2).GT.0.0).AND.(C(J1).GT.0.0))
     X                  A(J1,J2) = A(J1,J2)/SQRT(EVALS(M-J2)*C(J1))
              IF ((EVALS(M-J2).EQ.0.0).OR.(C(J1).EQ.0.0)) A(J1,J2)=0.0
  200      CONTINUE
  300   CONTINUE

        RETURN
        END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C
C   Output contributions of row-points to first NF factors.                  C
C----------------------------------------------------------------------------C
        SUBROUTINE COUTCX(N,M,PRJN,R,RCN,NF)
        REAL    PRJN(N,M), R(N), RCN(N,M)

        NUM = NF

        DO 300 J = 1, NUM
           COLSUM = 0.0
           DO 100 K = 1, N
              RCN(K,J) = R(K)*PRJN(K,J)**2
              COLSUM = COLSUM + RCN(K,J)
  100      CONTINUE
C          Normalize so that sum of contribns. for a factor equals 1
           DO 200 K = 1, N
              RCN(K,J) = RCN(K,J)/COLSUM
  200      CONTINUE
  300   CONTINUE

        RETURN
        END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C
C   Output contributions of column-points to first NF factors.               C
C----------------------------------------------------------------------------C
        SUBROUTINE COUTCY(M,PRJNS,C,CCN,NF)
        REAL    PRJNS(M,M), C(M), CCN(M,M)

        NUM = NF

        DO 300 J = 1, NUM
           COLSUM = 0.0
           DO 100 K = 1, M
               CCN(K,J) = C(K)*PRJNS(K,J)**2
               COLSUM = COLSUM + CCN(K,J)
  100      CONTINUE
C          Normalize so that contribns. to a factor sum to 1
           DO 200 K = 1, M
              CCN(K,J) = CCN(K,J)/COLSUM
  200      CONTINUE
  300   CONTINUE

        RETURN
        END




C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C
C                                                            C
C  HIERARCHICAL CLUSTERING using (user-specified) criterion. C
C                                                            C
C  Parameters:                                               C
C                                                            C
C  DATA(N,M)         input data matrix,                      C
C  DISS(LEN)         dissimilarities in lower half diagonal  C
C                    storage; LEN = N.N-1/2,                 C
C  IOPT              clustering criterion to be used,        C
C  IA, IB, CRIT      history of agglomerations; dimensions   C
C                    N, first N-1 locations only used,       C
C  MEMBR, NN, DISNN  vectors of length N, used to store      C 
C                    cluster cardinalities, current nearest  C
C                    neighbour, and the dissimilarity assoc. C
C                    with the latter.                        C
C  FLAG              boolean indicator of agglomerable obj./ C
C                    clusters.                               C
C                                                            C
C  F. Murtagh, ESA/ESO/STECF, Garching, February 1986.       C
C                                                            C
C------------------------------------------------------------C
      SUBROUTINE HC(N,M,LEN,IOPT,DATA,IA,IB,CRIT,MEMBR,NN,DISNN,
     X                FLAG,DISS)
      REAL DATA(N,M),MEMBR(N),DISS(LEN)
      INTEGER IA(N),IB(N)
      REAL CRIT(N)
      DIMENSION NN(N),DISNN(N)
      LOGICAL FLAG(N)
      REAL INF
      DATA INF/1.E+20/
C
C  Initializations
C
      DO I=1,N
         MEMBR(I)=1.
         FLAG(I)=.TRUE.
      ENDDO
      NCL=N
C
C  Construct dissimilarity matrix
C
      DO I=1,N-1
         DO J=I+1,N
            IND=IOFFSET(N,I,J)
            DISS(IND)=0.
            DO K=1,M
               DISS(IND)=DISS(IND)+(DATA(I,K)-DATA(J,K))**2
            ENDDO
            IF (IOPT.EQ.1) DISS(IND)=DISS(IND)/2.
C           (Above is done for the case of the min. var. method
C            where merging criteria are defined in terms of variances
C            rather than distances.)
          ENDDO
       ENDDO
C
C  Carry out an agglomeration - first create list of NNs
C
      DO I=1,N-1
         DMIN=INF
         DO J=I+1,N
            IND=IOFFSET(N,I,J)
            IF (DISS(IND).GE.DMIN) GOTO 500
               DMIN=DISS(IND)
               JM=J
  500    CONTINUE
         ENDDO
         NN(I)=JM
         DISNN(I)=DMIN
      ENDDO
C
  400 CONTINUE
C     Next, determine least diss. using list of NNs
      DMIN=INF
      DO I=1,N-1
         IF (.NOT.FLAG(I)) GOTO 600
         IF (DISNN(I).GE.DMIN) GOTO 600
            DMIN=DISNN(I)
            IM=I
            JM=NN(I)
  600    CONTINUE
      ENDDO
      NCL=NCL-1
C
C  This allows an agglomeration to be carried out.
C
      I2=MIN0(IM,JM)
      J2=MAX0(IM,JM)
      IA(N-NCL)=I2
      IB(N-NCL)=J2
      CRIT(N-NCL)=DMIN
C
C  Update dissimilarities from new cluster.
C
      FLAG(J2)=.FALSE.
      DMIN=INF
      DO K=1,N
         IF (.NOT.FLAG(K)) GOTO 800
         IF (K.EQ.I2) GOTO 800
         X=MEMBR(I2)+MEMBR(J2)+MEMBR(K)
         IF (I2.LT.K) THEN
                           IND1=IOFFSET(N,I2,K)
                      ELSE
                           IND1=IOFFSET(N,K,I2)
         ENDIF
         IF (J2.LT.K) THEN
                           IND2=IOFFSET(N,J2,K)
                      ELSE
                           IND2=IOFFSET(N,K,J2)
         ENDIF
         IND3=IOFFSET(N,I2,J2)
         XX=DISS(IND3)
C
C  WARD'S MINIMUM VARIANCE METHOD - IOPT=1.
C
         IF (IOPT.EQ.1) THEN
            DISS(IND1)=(MEMBR(I2)+MEMBR(K))*DISS(IND1)+
     X                 (MEMBR(J2)+MEMBR(K))*DISS(IND2)-
     X                 MEMBR(K)*XX
            DISS(IND1)=DISS(IND1)/X
         ENDIF
C
C  SINGLE LINK METHOD - IOPT=2.
C
         IF (IOPT.EQ.2) THEN
            DISS(IND1)=MIN(DISS(IND1),DISS(IND2))
         ENDIF
C
C  COMPLETE LINK METHOD - IOPT=3.
C
         IF (IOPT.EQ.3) THEN
            DISS(IND1)=MAX(DISS(IND1),DISS(IND2))
         ENDIF
C
C  AVERAGE LINK (OR GROUP AVERAGE) METHOD - IOPT=4.
C
         IF (IOPT.EQ.4) THEN
            DISS(IND1)=(MEMBR(I2)*DISS(IND1)+MEMBR(J2)*DISS(IND2))/
     X                 (MEMBR(I2)+MEMBR(J2))
         ENDIF
C
C  MCQUITTY'S METHOD - IOPT=5.
C
         IF (IOPT.EQ.5) THEN
            DISS(IND1)=0.5*DISS(IND1)+0.5*DISS(IND2)
         ENDIF
C
C  MEDIAN (GOWER'S) METHOD - IOPT=6.
C
         IF (IOPT.EQ.6) THEN
            DISS(IND1)=0.5*DISS(IND1)+0.5*DISS(IND2)-0.25*XX
         ENDIF
C
C  CENTROID METHOD - IOPT=7.
C
         IF (IOPT.EQ.7) THEN
            DISS(IND1)=(MEMBR(I2)*DISS(IND1)+MEMBR(J2)*DISS(IND2)-
     X          MEMBR(I2)*MEMBR(J2)*XX/(MEMBR(I2)+MEMBR(J2)))/
     X          (MEMBR(I2)+MEMBR(J2))
            ENDIF
C
         IF (I2.GT.K) GOTO 800
         IF (DISS(IND1).GE.DMIN) GOTO 800
            DMIN=DISS(IND1)
            JJ=K
  800    CONTINUE
      ENDDO
      MEMBR(I2)=MEMBR(I2)+MEMBR(J2)
      DISNN(I2)=DMIN
      NN(I2)=JJ
C
C  Update list of NNs insofar as this is required.
C
      DO I=1,N-1
         IF (.NOT.FLAG(I)) GOTO 900
         IF (NN(I).EQ.I2) GOTO 850
         IF (NN(I).EQ.J2) GOTO 850
         GOTO 900
  850    CONTINUE
C        (Redetermine NN of I:)
         DMIN=INF
         DO J=I+1,N
            IND=IOFFSET(N,I,J)
            IF (.NOT.FLAG(J)) GOTO 870
            IF (I.EQ.J) GOTO 870
            IF (DISS(IND).GE.DMIN) GOTO 870
               DMIN=DISS(IND)
               JJ=J
  870       CONTINUE
         ENDDO
         NN(I)=JJ
         DISNN(I)=DMIN
  900    CONTINUE
      ENDDO
C
C  Repeat previous steps until N-1 agglomerations carried out.
C
      IF (NCL.GT.1) GOTO 400
C
C
      RETURN
      END
C
C
      FUNCTION IOFFSET(N,I,J)
C  Map row I and column J of upper half diagonal symmetric matrix 
C  onto vector.
      IOFFSET=J+(I-1)*N-(I*(I+1))/2
      RETURN
      END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C
C                                                                C
C  HIERARCHICAL CLUSTERING using Minimum Variance Criterion,     C
C  using the O(N**2) time Nearest Neighbour Chain algorithm.     C
C                                                                C
C  Parameters:                                                   C
C                                                                C
C  DATA(N,M) :      input data,                                  C
C  IA(N), IB(N), CRIT(N) : sequence of agglomerands and          C
C                   values returned (only locations 1 to N-1     C
C                   are of interest),                            C
C  MEMBR(N), DISS(N), ICHAIN(N) : used in the routines to store  C
C                   cluster cardinalities, nearest neighbour     C    
C                   dissimilarities, and the NN-chain.           C
C  FLAG(N) :        (boolean) used to indicate agglomerable      C
C                   objects and clusters.                        C
C                                                                C
C  Reference: Murtagh, Multidimensional Clustering Algorithms,   C
C             Physica-Verlag, 1985.                              C
C                                                                C
C  F. Murtagh, ESA/ESO/STECF, Garching, February 1986.           C
C                                                                C
C----------------------------------------------------------------C
      SUBROUTINE HCON2(N,M,DATA,IA,IB,CRIT,MEMBR,DISS,ICHAIN,FLAG,istat)
      REAL    MEMBR(N), DATA(N,M), DISS(N), CRIT(N)
      INTEGER ICHAIN(N), IA(N), IB(N)
      REAL INF
      LOGICAL FLAG(N)
      DATA INF/1.E+25/
C     EQUIVALENCE (ICHAIN(1),IA(1)),(DISS(1),CRIT(1))
C
      DO 150 I=1,N
         MEMBR(I)=1
         FLAG(I)=.TRUE.
  150 CONTINUE
      NCL=N
      I1=1
C
C  Start the NN-chain:
C
  200 LEN=N
      ICHAIN(LEN)=I1
      DISS(LEN)=INF
C
C  Determine NN of object I1
C
  300 FLAG(I1)=.FALSE.
C
C  Turn off FLAG so that 0 diss. of I1 with self not obtained.
C
      D=DISS(LEN)
      IF (LEN.LT.N) I2=ICHAIN(LEN+1)
C
C  For identical diss.'s, above ensures that RNN will be found.
C
      CALL DETNN(DATA,FLAG,MEMBR,N,M,I1,I2,D)
      FLAG(I1)=.TRUE.
C
C  If LEN = 1 place obj. I2 as second obj. in NN-chain.
C
      IF (LEN.LT.N) GOTO 350
      LEN=LEN-1
      IF (LEN.LT.N-NCL) GOTO 700
      ICHAIN(LEN)=I2
      DISS(LEN)=D
      GOTO 500
C
C  If LEN < N distinguish between having RNN & continuing NN-chain.
C
  350 CONTINUE
      IF (I2.NE.ICHAIN(LEN+1)) GOTO 400
C
C  Have RNN.
C
      NCL=NCL-1
      CALL AGGLOM(I1,I2,D,DATA,MEMBR,FLAG,IA,IB,CRIT,NCL,N,M)
      LEN=LEN+2
      GOTO 500
  400 CONTINUE
C
C  Grow extra link on NN-chain.
C
      IDUM=ICHAIN(LEN+1)
      FLAG(IDUM)=.FALSE.
      LEN=LEN-1
      IF (LEN.LE.N-NCL) GOTO 700
      ICHAIN(LEN)=I2
      DISS(LEN)=D
      GOTO 500
C
C  Select obj. for continuing to grow (or restarting) NN-chain.
C
  500 CONTINUE
      IF (NCL.EQ.1) GOTO 600
      IF (LEN.EQ.N+1) GOTO 550
      I1=ICHAIN(LEN)
      FLAG(I1)=.TRUE.
      IDUM=ICHAIN(LEN+1)
      IF (LEN.LT.N) FLAG(IDUM)=.TRUE.
C
C  Reestablish agglomerability of objects in NN-chain.
C
      GOTO 300
  550 CALL NEXT(FLAG,I1,N)
      GOTO 200
C
  600 CONTINUE
      RETURN
C
C     SHOULD NEVER REACH HERE...
 700  CONTINUE
C  700 WRITE(6,750)
C  750 FORMAT(' ERROR IN NN-CHAIN ROUTINE - INSUFFICIENT CHAIN SPACE'/)
      istat = 1
      return
      END
C------------------------------------------------------------------------------
      SUBROUTINE DETNN(DATA,FLAG,MEM,N,M,I1,I2,D)
C
C  Determine a nearest neighbour.
C
      REAL DATA(N,M),MEM(N)
      LOGICAL FLAG(N)
C
      DO 200 I=1,N
         IF (.NOT.FLAG(I)) GOTO 200
         DISS=0.
         DO 100 J=1,M
  100    DISS=DISS+(DATA(I1,J)-DATA(I,J))*(DATA(I1,J)-DATA(I,J))
         DISS=DISS*MEM(I)*MEM(I1)/(MEM(I1)+MEM(I))
         IF (DISS.GE.D) GOTO 200
            D=DISS
            I2=I
  200 CONTINUE
C
      RETURN
      END
C------------------------------------------------------------------------------
      SUBROUTINE AGGLOM(I1,I2,D,DATA,MEM,FLAG,IA,IB,CRIT,NCL,N,M)
C
C  Carry out an agglomeration.
C
      REAL MEM(N),DATA(N,M),CRIT(N)
      INTEGER IA(N),IB(N)
      LOGICAL FLAG(N)
      INTEGER O1,O2,LB,UB
C
C
      O1=MIN0(I1,I2)
      O2=MAX0(I1,I2)
      DO 100 J=1,M
         DATA(O1,J)=( MEM(O1)*DATA(O1,J)+MEM(O2)*DATA(O2,J) )
     X            / (MEM(O1)+MEM(O2))
         DATA(O2,J)=DATA(O1,J)
  100 CONTINUE
      NAGGL=N-NCL
      MEM(O1)=MEM(O1)+MEM(O2)
      FLAG(O2)=.FALSE.
C
C  Keep sorted list of criterion values: find 1st where new crit. fits.
C
      I=NAGGL-1
  120 IF (D.GE.CRIT(I)) GOTO 140
      I=I-1
      IF (I.GE.1) GOTO 120
C
C  Arriving here must mean that D > all crit. values found so far.
C
      I=0
  140 CONTINUE
C
C  Now, shift rightwards from I+1 to AGGL-1 to make room for new crit.
C
      LB=I+1
      UB=NAGGL-1
      IF (LB.GT.UB) GOTO 180
      J=UB
  160 J1=J+1
      IA(J1)=IA(J)
      IB(J1)=IB(J)
      CRIT(J1)=CRIT(J)
      J=J-1
      IF (J.GE.LB) GOTO 160
  180 CONTINUE
      IA(LB)=O1
      IB(LB)=O2
      CRIT(LB)=D
C
      RETURN
      END
C------------------------------------------------------------------------------
      SUBROUTINE NEXT(FLAG,I1,N)
C
C  Determine next agglomerable object/cluster.
C
      LOGICAL FLAG(N)
C
      NXT=I1+1
      IF (NXT.GT.N) GOTO 150
      DO 100 I=NXT,N
         IF (FLAG(I)) GOTO 500
  100 CONTINUE
  150 DO 200 I=1,I1
         IF (FLAG(I)) GOTO 500
  200 CONTINUE
C
C     Should never get to here...
C     STOP
C
  500 I1=I
C
      RETURN
      END
C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C
C                                                               C
C  Given a HIERARCHIC CLUSTERING, described as a sequence of    C
C  agglomerations, prepare the seq. of aggloms. and "horiz."    C
C  order of objects for plotting the dendrogram using S routine C
C  'plclust'.                                                   C
C                                                               C
C  Parameters:                                                  C
C                                                               C
C  IA, IB:       vectors of dimension N defining the agglomer-  C
C                 ations.                                       C
C  IIA, IIB:     used to store IA and IB values differently     C
C                (in form needed for S command `plclust`        C
C  IORDER:       "horiz." order of objects for dendrogram       C
C                                                               C
C  F. Murtagh, ESA/ESO/STECF, Garching, June 1991               C
C                                                               C
C  HISTORY                                                      C
C                                                               C
C  Adapted from routine HCASS, which additionally determines    C
C   cluster assignments at all levels, at extra comput. expense C
C                                                               C
C---------------------------------------------------------------C
      SUBROUTINE HCASS2(N,IA,IB,IORDER,IIA,IIB)
      INTEGER IA(N),IB(N),IORDER(N),IIA(N),IIB(N)
C
C     Following bit is to get seq. of merges into format acceptable to plclust
C     I coded clusters as lowest seq. no. of constituents; S's `hclust' codes
C     singletons as -ve numbers, and non-singletons with their seq. nos.
C
      DO 912 I=1,N
         IIA(I)=IA(I)
         IIB(I)=IB(I)
  912 CONTINUE
      DO 915 I=1,N-2
C        In the following, smallest (+ve or -ve) seq. no. wanted
         K=MIN(IA(I),IB(I))
         DO 913 J=I+1, N-1
            IF(IA(J).EQ.K) IIA(J)=-I
            IF(IB(J).EQ.K) IIB(J)=-I
  913    CONTINUE
  915 CONTINUE
      DO 916 I=1,N-1
         IIA(I)=-IIA(I)
         IIB(I)=-IIB(I)
  916 CONTINUE
      DO 917 I=1,N-1
         IF (IIA(I).GT.0.AND.IIB(I).LT.0) THEN
            K = IIA(I)
            IIA(I) = IIB(I)
            IIB(I) = K
         ENDIF
         IF (IIA(I).GT.0.AND.IIB(I).GT.0) THEN
            K1 = MIN(IIA(I),IIB(I))
            K2 = MAX(IIA(I),IIB(I))
            IIA(I) = K1
            IIB(I) = K2
         ENDIF
  917 CONTINUE
C
C
C     NEW PART FOR `ORDER'
C
      IORDER(1) =IIA(N-1)
      IORDER(2) =IIB(N-1)
      LOC=2
      DO 175 I=N-2,1,-1
        DO 169 J=1,LOC
          IF(IORDER(J).EQ.I) THEN
C           REPLACE IORDER(J) WITH IIA(I) AND IIB(I)
            IORDER(J)=IIA(I)
            IF (J.EQ.LOC) THEN
                LOC=LOC+1
                IORDER(LOC)=IIB(I)
                GOTO 171
            ENDIF
            LOC=LOC+1
            DO 95 K=LOC,J+2,-1
               IORDER(K)=IORDER(K-1)
  95        CONTINUE
            IORDER(J+1)=IIB(I)
            GOTO 171
          ENDIF
 169    CONTINUE
C       SHOULD NEVER REACH HERE
 171    CONTINUE
 175  CONTINUE
C 
C
C
      DO 181 I=1,N
         IORDER(I) = -IORDER(I)
 181  CONTINUE
C
C
      RETURN
      END




C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C  HIERARCHICAL CLUSTERING in M-DIMENSIONAL SPACE. F.Murtagh, Sep. '91.
C  Clusters are determined at each agglomeration, so that they can be graphed.
C
C  Data/variables for all routines:
C
C  DATA(N,M)         input data 
C  DAT2(N*M)         copy of part of DATA (i.e. rows corresp. to a cluster).
C                    In vector form.
C  N,M               nos. of rows and cols. assoc. with input data matrix,
C  IA, IB, CRIT      history of aggloms.; subtree sequence numbers in IA, IB;
C                    the seq. no. of a cluster is the smaller seq. no. of the
C                    agglomerands; CRIT contains criterion value at which
C                    agglomeration takes place; dimsensions of these vectors:
C                    N, of which the first N-1 locations, only, are used.    
C                    With this notation, a test for a non-singleton cluster 
C                    is: a cluster seq. no. must additionally exist at a lower
C                    index location in the IA/IB lists.  This lower index 
C                    location is assocated with a subcluster, or a singleton.
C  MEMBR             vector of length N, used to store cluster cardinalities;
C                    initially all values are 1, indicating that each seq. no.
C                    refers to a singleton cluster.  MEMBR is a real vector,
C                    although storing integer values at all times, for comput-
C                    ational convenience.
C  FLAG              boolean indicator of agglomerable objects/clusters.  Note
C                    that since seq. no. can refer to both singleton-object 
C                    and to non-singleton cluster, FLAG is necessary. 
C  IKLASS            1 (resp. 0) in row i, col. j indicates that object j is
C                    present (resp. absent) in cluster i.
C                    N-1 rows only used. Rows <-> clusters, cols. <-> objects.
C  POTCL             potential cluster: N-length vector of 1s and 0s.
C  POTCL1, POTCL2    agglomerable clusters: each of these is an N-length 
C                    integer-valued boolean indicator vector.
C  ARRAY1(M,M), ARRAY2(M,M), VECT1(M), VECT2(M)   Work arrays and vectors, used
C                    to store the covariance matrix and for other purposes by
C                    the principal components analysis routine, PRCOAN.
C  CENTR(M)          vector of centers of a cluster, used by PRCOAN.
C  V1, V2, VTOT      real scalars used to store variance values.
C
C  Subroutines
C
C  INIT    Initialize
C  GBD     Get best dissimilarity (min. var. criterion, with alpha, version)
C  GBD2    Get best dissimilarity (all other agglomerative criteria)
C  AGG     Agglomerate
C  GNCM    Get new cluster memberships
C  CM      Cluster memberships 
C  AL      Assess linearity (min. var. criterion, with alpha, version)
C  AL2     Assess linearity (all other agglomerative criteria: actually this
C         routine does nothing more than determine memberships of agglomerands)
C  CLLIN   CLuster LINearity: driver for principal components routine
C  PRCOAN  Determines covariance matrix, and coordinates SVD
C  LPCOVCL, LPTRED2, LPTQL2  Used by PRCOAN
C------------------------------------------------------------------------------
C
C
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C  Initializations
C------------------------------------------------------------------------------
      SUBROUTINE        INIT(MEMBR,FLAG,IKLASS,N,M)
      REAL              MEMBR(N)
      INTEGER           IKLASS(N,N)
      LOGICAL           FLAG(N)
C
      DO 20 I = 1,N
         MEMBR(I) = 1.
         FLAG(I)  = .TRUE.
         DO 10 J = 1,N-1
            IKLASS(J,I) = 0
 10      CONTINUE
 20   CONTINUE
C
      RETURN
      END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C     Determine least diss. using original data.  GBD: Get Best Dissimilarity.
C------------------------------------------------------------------------------
      SUBROUTINE GBD
     .       (MEMBR,FLAG,DLEAST,ILEAST,JLEAST,NCL,IA,IB,IKLASS,POTCL,
     .        array1,array2,vect1,vect2,dat2,N,M,ALPHA,DATA,CENTR)
      LOGICAL    FLAG(N)
      REAL       MEMBR(N),DATA(N,M),INF,CENTR(M)
      REAL       array1(m,m),array2(m,m),vect1(m),vect2(m),dat2(1)
      INTEGER    IA(N), IB(N), IKLASS(N,N), POTCL(N)
C
      INF     = 1.E+20
      DLEAST  = INF
      DO 30 I = 1, N
C-- FLAG controls whether or not seq. no. refers to still active obj./clust.
         IF (.NOT.FLAG(I)) GOTO 700
         DO 20 J = 1,N
            IF (.NOT.FLAG(J)) GOTO 600
            IF (I.EQ.J) GOTO 600
C-- ILAGG and JLAGG - pair of last agglomerated objects/clusters:
            ILAGG = 0
            JLAGG = 0
C-- In the case of clusters of > 2 members, get the member seq. nos.:
C-- Subroutine CM: Cluster Memberships
            IF (MEMBR(I).GT.1) CALL CM(I,ILAGG,NCL,IA,IB,N)
            IF (MEMBR(J).GT.1) CALL CM(J,JLAGG,NCL,IA,IB,N)
C-- Variances of a potential 2-member cluster:
            IF (MEMBR(I).EQ.1.AND.MEMBR(J).EQ.1) THEN
               DISSIM = 0.0
               DO 10 K = 1, M
                  DISSIM = DISSIM + 
     .              (DATA(I,K)-DATA(J,K))*(DATA(I,K)-DATA(J,K))
 10            CONTINUE
               V1   = 0.5*DISSIM
               V2   = 0.0
               VTOT = V1
            ENDIF
C-- Variances of a cluster with potentially > 2 members: 
C-- Subroutine AL: Assess Linearity
            IF (MEMBR(I).GT.1.OR.MEMBR(J).GT.1) THEN
             CALL AL(I,J,ILAGG,JLAGG,IKLASS,POTCL,
     .       array1,array2,vect1,vect2,dat2,V1,V2,VTOT,N,M,DATA,CENTR)
            ENDIF
C-- ALPHA is the linearity vs. compactness coeff. Cf. Murtagh & Raftery, '84.
            XCRIT = ALPHA*V1 + (VTOT-V1)
            IF (XCRIT.GE.DLEAST) GOTO 600
               DLEAST = XCRIT
               ILEAST = I
               JLEAST = J
  600       CONTINUE
   20      CONTINUE
  700    CONTINUE
   30 CONTINUE
C
      RETURN
      END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C     Determine least diss. using original data.  GBD: Get Best Dissimilarity.
C     GBD2 -- for all criteria except min. var. one.
C     (In min. var. criterion, allow for linearity-determining; not in other.)
C------------------------------------------------------------------------------
      SUBROUTINE GBD2
     .     (MEMBR,FLAG,DLEAST,ILEAST,JLEAST,NCL,IA,IB,IKLASS,POTCL1,
     .      POTCL2,array1,array2,vect1,vect2,dat2,N,M,DATA,METHOD)
      LOGICAL    FLAG(N)
      REAL       MEMBR(N),DATA(N,M),INF
      REAL       array1(m,m),array2(m,m),vect1(m),vect2(m),dat2(1)
      INTEGER    IA(N), IB(N), IKLASS(N,N), POTCL1(N), POTCL2(N)
C
      INF     = 1.E+20
      DLEAST  = INF
      DO 30 I = 1, N
C-- FLAG controls whether or not seq. no. refers to still active obj./clust.
         IF (.NOT.FLAG(I)) GOTO 700
         DO 20 J = 1,N
            IF (.NOT.FLAG(J)) GOTO 600
            IF (I.EQ.J) GOTO 600
C-- ILAGG and JLAGG - pair of last agglomerated objects/clusters:
            ILAGG = 0
            JLAGG = 0
C-- In the case of clusters of > 2 members, get the member seq. nos.:
C-- Subroutine CM: Cluster Memberships
            IF (MEMBR(I).GT.1) CALL CM(I,ILAGG,NCL,IA,IB,N)
            IF (MEMBR(J).GT.1) CALL CM(J,JLAGG,NCL,IA,IB,N)
            DMIN = 1000000.0
C-- Variances of a potential 2-member cluster:
            IF (MEMBR(I).EQ.1.AND.MEMBR(J).EQ.1) THEN
               DISSIM = 0.0
               DO 10 K = 1, M
                  DISSIM = DISSIM + 
     .              (DATA(I,K)-DATA(J,K))*(DATA(I,K)-DATA(J,K))
 10            CONTINUE
               DMIN = DISSIM
            ENDIF
            DISMIN = 1000000.0
C-- Variances of a cluster with potentially > 2 members: 
C-- Subroutine AL2: Modeled on `Assess Linearity' routine (AL), but in this
C-- we firstly only det. the clusters' members
            IF (MEMBR(I).GT.1.OR.MEMBR(J).GT.1) THEN
                CALL AL2(I,J,ILAGG,JLAGG,IKLASS,POTCL1,POTCL2,
     .          array1,array2,vect1,vect2,dat2,N,M,DATA)
C-- Now POTCL1 has 1 values to indicate members of cluter 1; ditto for POTCL2.
                IF (METHOD.EQ.2) THEN
C--                Single link
                   DISMIN = 100000.0
                   DO 400 I1 = 1, N
                      IF (POTCL1(I1).EQ.1) THEN
                         DO 380 I2 = 1, N
                            IF (POTCL2(I2).EQ.1) THEN
                               DISSIM = 0.0
                               DO 360 K2 = 1, M
                                  DISSIM=DISSIM+
     .              (DATA(I1,K2)-DATA(I2,K2))*(DATA(I1,K2)-DATA(I2,K2))
 360                           CONTINUE
                               IF (DISSIM.LT.DISMIN) DISMIN = DISSIM
                            ENDIF
 380                     CONTINUE
                      ENDIF
 400               CONTINUE
                ENDIF
C--
                IF (METHOD.EQ.3) THEN
C--                Complete link
                   DISMIN = 0.0
                   DO 500 I1 = 1, N
                      IF (POTCL1(I1).EQ.1) THEN
                         DO 480 I2 = 1, N
                            IF (POTCL2(I2).EQ.1) THEN
                               DISSIM = 0.0
                               DO 460 K2 = 1, M
                                  DISSIM=DISSIM+
     .              (DATA(I1,K2)-DATA(I2,K2))*(DATA(I1,K2)-DATA(I2,K2))
 460                           CONTINUE
                               IF (DISSIM.GT.DISMIN) DISMIN = DISSIM
                            ENDIF
 480                     CONTINUE
                      ENDIF
 500               CONTINUE
                ENDIF
C--
                IF (METHOD.EQ.4) THEN
C--                Average link
                   DISMIN = 0.0
                   DISSIM = 0.0
                   INLINKS = 0
                   DO 590 I1 = 1, N
                      IF (POTCL1(I1).EQ.1) THEN
                         DO 580 I2 = 1, N
                            IF (POTCL2(I2).EQ.1) THEN
                               INLINKS = INLINKS + 1
                               DO 560 K2 = 1, M
                                  DISSIM=DISSIM+
     .              (DATA(I1,K2)-DATA(I2,K2))*(DATA(I1,K2)-DATA(I2,K2))
 560                           CONTINUE
                            ENDIF
 580                     CONTINUE
                      ENDIF
 590                  CONTINUE
                      DISMIN = DISSIM/FLOAT(INLINKS)
                ENDIF
C--
            ENDIF
C
            
            XCRIT = MIN(DMIN,DISMIN)
            IF (XCRIT.GE.DLEAST) GOTO 600
               DLEAST = XCRIT
               ILEAST = I
               JLEAST = J
  600       CONTINUE
   20      CONTINUE
  700    CONTINUE
   30 CONTINUE
C
      RETURN
      END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C     Carry out an agglomeration
C------------------------------------------------------------------------------
      SUBROUTINE AGG(ILEAST,JLEAST,DLEAST,NCL,IA,IB,CRIT,MEMBR,FLAG,N)
      INTEGER    IA(N), IB(N)
      REAL       CRIT(N), MEMBR(N)
      LOGICAL    FLAG(N)
C
      I           = MIN0(ILEAST,JLEAST)
      J           = MAX0(ILEAST,JLEAST)
      ILEAST      = I
      JLEAST      = J
      IA(N-NCL)   = I
      IB(N-NCL)   = J
      CRIT(N-NCL) = DLEAST
C
      RETURN
      END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C  Update cluster memberships array, iklass
C  GNCM: Get New Cluster Memberships
C------------------------------------------------------------------------------
      SUBROUTINE GNCM(I2,J2,NCL,IA,IB,MEMBR,IKLASS,FLAG,N)
      INTEGER    IA(N), IB(N), IKLASS(N,N)
      REAL       MEMBR(N)
      LOGICAL    FLAG(N)
C
C     We are dealing with clusters I2 and J2.  Are these multiobject clusters,
C     or singletons?  (Remember, the seq. no. values contained in I2 and J2 
C     could be either.)  If I2 (resp. J2) is a singleton, then the value of
C     I2 (resp. J2) is the one member.  If I2 (resp. J2) is a non-singleton
C     cluster, then we must pick up the most recently updated row of array
C     IKLASS.  1 and 0 values in this row indicate memberships in the cluster.
C
C     Use cardinalities vector, MEMBR, to find whether we have a singleton or
C     a non-singleton cluster.
      IAINE = MEMBR(I2)
      IBENJ = MEMBR(J2)
C      
      IF (IAINE.GT.1) GOTO 610
C        Singleton, i.e. one object with seq. no. I2 
         IKLASS(N-NCL,I2)=1
         GOTO 620
  610    CONTINUE
C        Non-singleton, i.e. multi-object cluster
         CALL CM(I2, ILAGG, NCL,IA,IB,N)
         DO 10 ICLST = 1,N
            IF (IKLASS(ILAGG,ICLST).EQ.1) IKLASS(N-NCL,ICLST)=1
   10    CONTINUE
  620 CONTINUE
C
      IF (IBENJ.GT.1) GOTO 630
C        Singleton, i.e. one object with seq. no. J2
         IKLASS(N-NCL,J2)=1
         GOTO 640
  630    CONTINUE
C        Non-singleton cluster
         CALL CM(J2, JLAGG, NCL,IA,IB,N)
         DO 20 ICLST = 1,N
            IF (IKLASS(JLAGG,ICLST).EQ.1) IKLASS(N-NCL,ICLST)=1
   20    CONTINUE
  640 CONTINUE
C
C--  Finally, update MEMBR and FLAG:
      MEMBR(I2)    = MEMBR(I2) + MEMBR(J2)
      FLAG(J2)     = .FALSE.
C
      RETURN
      END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C     Given the seq. no. (identifier) of a non-singleton cluster, find where
C     it was most recently met with in the sequence of agglomerations.  The
C     agglomeration number will then provide us with the row of array IKLASS
C     (not used in this subroutine!) which in turn allows us to read off the
C     cluster's members.
C     ICL is the cluster seq. no., ILAGG is the value sought, and 
C     IA and IB relate to the sequence of agglomerations.
C     Subroutine CM: Cluster Memberships
C------------------------------------------------------------------------------
      SUBROUTINE CM(ICL, ILAGG, NCL, IA, IB,N)
      INTEGER    IA(N), IB(N)
C
C     What agglom was ICL last met at?
      NAGG = N-NCL-1
      DO 5 I = NAGG,1,-1
         IF (IA(I).EQ.ICL.OR.IB(I).EQ.ICL) GOTO 10
   5  CONTINUE
C--   Should never get to here.
   10 CONTINUE
      ILAGG = I
C
      RETURN
      END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C Subroutine AL: Assess Linearity
C Currently uses principal components analysis
C------------------------------------------------------------------------------
      SUBROUTINE AL(I,J,ILAST,JLAST,IKLASS,POTCL,
     .           array1,array2,vec1,vec2,dat2,V1,V2,VTOT,N,M,DATA,CENTR)
      REAL       array1(m,m),array2(m,m),vec1(m),vec2(m),dat2(1)
      REAL       DATA(N,M),CENTR(M)
      INTEGER    IKLASS(N,N), POTCL(N)
C
      DO 10 IOBJ = 1, N
         POTCL(IOBJ) = 0
 10   CONTINUE
C
C     J is NN of I.  If I is singleton, then ILAST is 0; ditto for J, JLAST.
      IF (ILAST.EQ.0) THEN 
            POTCL(I) = 1
         ELSE 
            DO 20 IOBJ = 1, N
               IF (IKLASS(ILAST,IOBJ).EQ.1) POTCL(IOBJ) = 1
 20         CONTINUE
      ENDIF
      IF (JLAST.EQ.0) THEN 
            POTCL(J) = 1
         ELSE 
            DO 30 IOBJ = 1, N
               IF (IKLASS(JLAST,IOBJ).EQ.1) POTCL(IOBJ) = 1
 30         CONTINUE
      ENDIF
C
C--   Subroutine CLLIN: CLuster LINearity
      CALL CLLIN(POTCL,array1,array2,vec1,vec2,dat2,V1,V2,VTOT,N,M,
     .            DATA,CENTR)
C
      RETURN
      END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C Subroutine AL2: Modeled on AL, Assess Linearity
C Det. clusters' memberships
C------------------------------------------------------------------------------
      SUBROUTINE AL2(I,J,ILAST,JLAST,IKLASS,POTCL1, POTCL2,
     .           array1,array2,vec1,vec2,dat2,N,M,DATA)
      REAL       array1(m,m),array2(m,m),vec1(m),vec2(m),dat2(1)
      REAL       DATA(N,M)
      INTEGER    IKLASS(N,N), POTCL1(N), POTCL2(N)
C
      DO 10 IOBJ = 1, N
         POTCL1(IOBJ) = 0
         POTCL2(IOBJ) = 0
 10   CONTINUE
C
C     J is NN of I.  If I is singleton, then ILAST is 0; ditto for J, JLAST.
      IF (ILAST.EQ.0) THEN 
            POTCL1(I) = 1
         ELSE 
            DO 20 IOBJ = 1, N
               IF (IKLASS(ILAST,IOBJ).EQ.1) POTCL1(IOBJ) = 1
 20         CONTINUE
      ENDIF
      IF (JLAST.EQ.0) THEN 
            POTCL2(J) = 1
         ELSE 
            DO 30 IOBJ = 1, N
               IF (IKLASS(JLAST,IOBJ).EQ.1) POTCL2(IOBJ) = 1
 30         CONTINUE
      ENDIF
C
      RETURN
      END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C Subroutine CLLIN: CLuster LINearity
C Uses principal components analysis
C------------------------------------------------------------------------------
        SUBROUTINE CLLIN(POTCL,array1,array2,vect1,vect2,dat2,V1,V2,
     .                   VTOT,N,M,DATA,CENTR)
        INTEGER    POTCL(N)
        REAL       DATA(N,M),CENTR(M)
        REAL       array1(m,m),array2(m,m),vect1(m),vect2(m),dat2(1)
C
        NROW = 0
        DO 20 I = 1, N
           IF (POTCL(I).EQ.1) THEN
              NROW = NROW + 1
              DO 10 J = 1, M
c                 dat2(NROW,J) = DATA(I,J)
                  dat2(j+(nrow-1)*m) = data(i,j)
  10          CONTINUE
           ENDIF
  20    CONTINUE
C
	METHOD = 2
	IPRINT = 0
C       PCA follows.  Funny name, to distinguish it from other slightly
C       modified replications which might be used in the same S session
	CALL PRCOAN(N,NROW,M,DAT2,METHOD,IPRINT,ARRAY1,VECT1,VECT2,
     .              ARRAY2,IERR,CENTR)
             
C	IF (IERR.NE.0) GOTO 9000
C	GOTO 9900
C 9000	WRITE (6,*) ' ABNORMAL END: IERR =', IERR
 9900	CONTINUE
C
C-- For info: eigenvalues are in VECT1(M-J+1) for J=1,2,...min(NROW,M,7)
C-- 1st e-value is identical to sum of squared values in DAT2(*,1),
C-- 2nd e-value is identical to sum of squared values in DAT2(*,2), etc.
C-- Variance along 1st axis
        V1 = VECT1(M)               
C-- Variance along 2nd axis
        V2 = VECT1(M-1)             
C-- Total variance, - or at least 1st 7 variances
        VTOT = 0.0
        DO 40 J = 1, MIN(NROW,M,7)
           VTOT = VTOT + VECT1(M-J+1)
 40     CONTINUE
C

        RETURN
	END
C
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C  Carry out a PRINCIPAL COMPONENTS ANALYSIS (SVD, KARHUNEN-LOEVE EXPANSION).  
C  To call:   CALL PRCOAN(N,NROW,M,DATA,METHOD,IPRINT,A1,W1,W2,A2,IERR) 
C             where
C                                
C  N,NROW,M : integer dims. (NROW is actual # rows in use, out of N) of ...
C  DATA     : input data.              
C             On output, DATA contains in first 7 columns the projections of 
C             the row-points on the first 7 principal components. 
C  A1       : variance-covariance matrix, dimensions M * M.            
C             On output, A1 contains in the first 7 columns the projections of 
C             the column-points on the first 7 principal components. 
C  W1,W2    : real vectors of dimension M (see called routines for use).
C             On output, W1 contains the eigenvalues (read with index M-J+1, 
C             where running index J is J=1,7)
C  A2       : real array of dimensions M * M (see routines for use).   
C  IERR     : error indicator (normally zero).            
C             If IERR > 0, then its value indicates the eigenvalue for which
C             no convergence was obtained.      
C                                                      
C  Inputs here are N, M, DATA, (and IERR).
C                                 
C  HISTORY
C  Adapted from PCA routine written by F. Murtagh in 1986 by F. Murtagh in 
C  October 1990.     
C------------------------------------------------------------------------------
        SUBROUTINE PRCOAN(N,NROW,M,DATA,METHOD,IPRINT,A,
     .                    W,FV1,Z,IERR,CENTR)
        REAL    DATA(1), A(M,M), W(M), FV1(M), Z(M,M), CENTR(M)
C
C-- Form covariance matrix.
        CALL LPCOVCL(N,M,NROW,DATA,CENTR,A)
C
C-- Carry out eigenreduction.
        M2 = M
        CALL LPTRED2(M,M2,A,W,FV1,Z)
        CALL LPTQL2(M,M2,W,FV1,Z,IERR)
C
        RETURN  
        END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C  Det. covariances of cols. First det. means of cols., storing in WORK.
C------------------------------------------------------------------------------
        SUBROUTINE    LPCOVCL(N,M,NROW,DATA,WORK,OUT)
        DIMENSION     DATA(1), OUT(M,M), WORK(M)
C
        DO 30 J = 1, M
           WORK(J) = 0.0
           DO 20 I = 1, NROW
               WORK(J) = WORK(J) + DATA(J+(I-1)*M)
   20      CONTINUE
           WORK(J) = WORK(J)/FLOAT(NROW)
   30   CONTINUE
C
C-- Now centre the column points.
        DO 50 I = 1, NROW
           DO 40 J = 1, M
               DATA(J+(I-1)*M) = DATA(J+(I-1)*M)-WORK(J)
   40      CONTINUE
   50   CONTINUE
C
C-- Finally calculate the cross product matrix of the redefined data matrix.
        DO 80 J1 = 1, M
           DO 70 J2 = J1, M
              OUT(J1,J2) = 0.0
              DO 60 I = 1, NROW
                  OUT(J1,J2) = OUT(J1,J2) + 
     .                         DATA(J1+(I-1)*M)*DATA(J2+(I-1)*M)
   60         CONTINUE
              OUT(J2,J1) = OUT(J1,J2)
   70      CONTINUE
   80   CONTINUE
C
        RETURN
        END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C Reduce a real, symmetric matrix to a symmetric, tridiagonal matrix. 
C To call:    CALL LPTRED2(NM,N,A,D,E,Z)    where
C                                     
C NM = row dimension of A and Z;      
C N  = order of matrix A (will always be <= NM);
C A  = symmetric matrix of order N to be reduced to tridiag. form;
C D  = vector of dim. N containing, on output, diagonal elts. of
C      tridiagonal matrix.               
C E  = working vector of dim. at least N-1 to contain subdiagonal
C      elements.                          
C Z  = matrix of dims. NM by N containing, on output, orthogonal
C      transformation matrix producing the reduction. 
C                                            
C Normally a call to PTQL2 will follow the call to PTRED2 in order to
C produce all eigenvectors and eigenvalues of matrix A.
C                                        
C Algorithm used: Martin et al., Num. Math. 11, 181-195, 1968. 
C Reference: Smith et al., Matrix Eigensystem Routines - EISPACK
C Guide, Lecture Notes in Computer Science 6, Springer-Verlag, 
C 1976, pp. 489-494.                     
C-----------------------------------------------------------------------------
        SUBROUTINE LPTRED2(NM,N,A,D,E,Z)
        REAL A(NM,N),D(N),E(N),Z(NM,N)
C
        DO 100 I = 1, N
           DO 100 J = 1, I
              Z(I,J) = A(I,J)
  100   CONTINUE
        IF (N.EQ.1) GOTO 320
        DO 300 II = 2, N
           I = N + 2 - II
           L = I - 1
           H = 0.0
           SCALE = 0.0
           IF (L.LT.2) GOTO 130
           DO 120 K = 1, L
              SCALE = SCALE + ABS(Z(I,K))
  120      CONTINUE
           IF (SCALE.NE.0.0) GOTO 140
  130      E(I) = Z(I,L)
           GOTO 290
  140      DO 150 K = 1, L
              Z(I,K) = Z(I,K)/SCALE
              H = H + Z(I,K)*Z(I,K)
  150      CONTINUE
C
           F = Z(I,L)
           G = -SIGN(SQRT(H),F)
           E(I) = SCALE * G
           H = H - F * G
           Z(I,L) = F - G
           F = 0.0
C
           DO 240 J = 1, L
              Z(J,I) = Z(I,J)/H
              G = 0.0
C             Form element of A*U.
              DO 180 K = 1, J
                 G = G + Z(J,K)*Z(I,K)
  180         CONTINUE
              JP1 = J + 1
              IF (L.LT.JP1) GOTO 220
              DO 200 K = JP1, L
                 G = G + Z(K,J)*Z(I,K)
  200         CONTINUE
C             Form element of P where P = I - U U' / H .
  220         E(J) = G/H
              F = F + E(J) * Z(I,J)
  240      CONTINUE
           HH = F/(H + H)
C          Form reduced A.
           DO 260 J = 1, L
              F = Z(I,J)
              G = E(J) - HH * F
              E(J) = G
              DO 250 K = 1, J
                 Z(J,K) = Z(J,K) - F*E(K) - G*Z(I,K)
  250         CONTINUE
  260      CONTINUE
  290      D(I) = H
  300   CONTINUE
  320   D(1) = 0.0
        E(1) = 0.0
C       Accumulation of transformation matrices.
        DO 500 I = 1, N
           L = I - 1
           IF (D(I).EQ.0.0) GOTO 380
           DO 360 J = 1, L
              G = 0.0
              DO 340 K = 1, L
                 G = G + Z(I,K) * Z(K,J)
  340         CONTINUE
              DO 350 K = 1, L
                 Z(K,J) = Z(K,J) - G * Z(K,I)
  350         CONTINUE
  360      CONTINUE
  380      D(I) = Z(I,I)
           Z(I,I) = 1.0
           IF (L.LT.1) GOTO 500
           DO 400 J = 1, L
              Z(I,J) = 0.0
              Z(J,I) = 0.0
  400      CONTINUE
  500   CONTINUE
C
        RETURN
        END
C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C Determine eigenvalues and eigenvectors of a symmetric, tridiagonal matrix
C To call:    CALL LPTQL2(NM,N,D,E,Z,IERR)    where
C                             
C NM   = row dimension of Z;    
C N    = order of matrix Z;      
C D    = vector of dim. N containing, on output, eigenvalues;
C E    = working vector of dim. at least N-1;      
C Z    = matrix of dims. NM by N containing, on output, eigenvectors;
C IERR = error, normally 0, but 1 if no convergence.    
C                      
C Normally the call to PTQL2 will be preceded by a call to PTRED2 in 
C order to set up the tridiagonal matrix.   
C                     
C Algorithm used: QL method of Bowdler et al., Num. Math. 11,
C 293-306, 1968.                   
C Reference: Smith et al., Matrix Eigensystem Routines - EISPACK 
C Guide, Lecture Notes in Computer Science 6, Springer-Verlag,
C 1976, pp. 468-474.                 
C-----------------------------------------------------------------------------
        SUBROUTINE LPTQL2(NM,N,D,E,Z,IERR)
        REAL    D(N), E(N), Z(NM,N)
        DATA    EPS/1.E-12/
C
        IERR = 0
        IF (N.EQ.1) GOTO 1001
        DO 100 I = 2, N
           E(I-1) = E(I)
  100   CONTINUE
        F = 0.0
        B = 0.0
        E(N) = 0.0
C
        DO 240 L = 1, N
           J = 0
           H = EPS * (ABS(D(L)) + ABS(E(L)))
           IF (B.LT.H) B = H
C          Look for small sub-diagonal element.
           DO 110 M = L, N
              IF (ABS(E(M)).LE.B) GOTO 120
C             E(N) is always 0, so there is no exit through
C             the bottom of the loop.
  110      CONTINUE
  120      IF (M.EQ.L) GOTO 220
  130      IF (J.EQ.30) GOTO 1000
           J = J + 1
C          Form shift.
           L1 = L + 1
           G = D(L)
           P = (D(L1)-G)/(2.0*E(L))
           R = SQRT(P*P+1.0)
           D(L) = E(L)/(P+SIGN(R,P))
           H = G-D(L)
C
           DO 140 I = L1, N
              D(I) = D(I) - H
  140      CONTINUE
C
           F = F + H
C          QL transformation.
           P = D(M)
           C = 1.0
           S = 0.0
           MML = M - L
C
           DO 200 II = 1, MML
              I = M - II
              G = C * E(I)
              H = C * P
              IF (ABS(P).LT.ABS(E(I))) GOTO 150
              C = E(I)/P
              R = SQRT(C*C+1.0)
              E(I+1) = S * P * R
              S = C/R
              C = 1.0/R
              GOTO 160
  150         C = P/E(I)
              R = SQRT(C*C+1.0)
              E(I+1) = S * E(I) * R
              S = 1.0/R
              C = C * S
  160         P = C * D(I) - S * G
              D(I+1) = H + S * (C * G + S * D(I))
C             Form vector.
              DO 180 K = 1, N
                 H = Z(K,I+1)
                 Z(K,I+1) = S * Z(K,I) + C * H
                 Z(K,I) = C * Z(K,I) - S * H
  180         CONTINUE
  200      CONTINUE
           E(L) = S * P
           D(L) = C * P
           IF (ABS(E(L)).GT.B) GOTO 130
  220      D(L) = D(L) + F
  240   CONTINUE
C
C       Order eigenvectors and eigenvalues.
        DO 300 II = 2, N
           I = II - 1
           K = I
           P = D(I)
           DO 260 J = II, N
              IF (D(J).GE.P) GOTO 260
              K = J
              P = D(J)
  260      CONTINUE
           IF (K.EQ.I) GOTO 300
           D(K) = D(I)
           D(I) = P
           DO 280 J = 1, N
              P = Z(J,I)
              Z(J,I) = Z(J,K)
              Z(J,K) = P
  280      CONTINUE
  300   CONTINUE
C
        GOTO 1001
C       Set error - no convergence after 30 iterns.
 1000   IERR = L
 1001   RETURN
        END  
C-----------------------------------------------------------------------------









C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C
C                                                               C
C  Given a HIERARCHIC CLUSTERING, described as a sequence of    C
C  agglomerations, derive the assignments into clusters for     C
C  all (except the first and last) levels of the hierarchy.     C
C                                                               C
C  PARAMETERS                                                   C
C                                                               C
C  IA, IB:       vectors of dimension N defining the agglomer-  C
C                 ations.                                       C
C  HVALS:        vector of dim. N, used internally only.        C
C  ICLASS:       array of cluster assignments; dim. N by LEV.   C
C  IIA, IIB:     alt. coding of info. in IA and IB.             C
C                                                               C
C  F. Murtagh, Feb. 1986.                                       C
C                                                               C
C  HISTORY                                                      C
C                                                               C
C  Derived from routine HCASS.F, June 1991.    F.M.             C
C                                                               C
C---------------------------------------------------------------C
      SUBROUTINE ASSGN(N,NPLUS1,IA,IB,ICLASS,HVALS,IIA,IIB)
      INTEGER IA(N),IB(N),ICLASS(NPLUS1,N),HVALS(N),IIA(N),IIB(N)
C
C     1 - CONVERT CODING OF SEQ. OF AGGLOMS. FROM S-CONVENTION TO MINE
C     S-CONVENTION: -VE VAL. => OBS. SEQ. NO.; +VE VAL. => CLUSTER REF.
C     MY CONVENTION: ALL VALS. +VE.  IF VAL. ALREADY REFERENCED, THEN 
C     CLUSTER; COL.1 VALS. ALWAYS < COL.2 VALS.
c
c     1st step: for each values, flip in sign
c     Note that n should be # obs. minus 1
      do 100 i = 1, n
         iia(i) = -ia(i)
         iib(i) = -ib(i)
 100  continue
c
c     2nd step: leave references to observations alone (these are now +ve)
c     Clusters: seek 
      do 500 i = 1, n
         if (iia(i).lt.0) then
c           Because it is -ve, iia(i) is a seq. no. of a cluster.
c           Find this cluster, and get the lowest seq. no. of obs. in it.
            iclust = -iia(i)
c           lowest val. among iia(iclust) and iib(iclust)
            iia(i) = iia(iclust)
         endif
c        Same for iib(i)
         if (iib(i).lt.0) then
            iclust = -iib(i)
            iib(i) = iia(iclust)
         endif
c        Ensure that iia(i) < iib(i)
         if (iia(i).gt.iib(i)) then
            itemp = iia(i)
            iia(i) = iib(i)
            iib(i) = itemp
         endif
 500  continue
C
C  Pick out the clusters which the N objects belong to,
C  at levels N-2, N-3, ... 1 of the hierarchy.
C  The clusters are identified by the lowest seq. no. of
C  their members.
C  There are 2, 3, ... N-1 clusters, respectively, for the
C  above levels of the hierarchy.
C
c     NOTE: in following N = # aggloms.; # obs. = N+1
      HVALS(1)=1
      HVALS(2)=IIB(N)
      LOC=3
      DO 59 I=N-1,1,-1
         DO 52 J=1,LOC-1
            IF (IIA(I).EQ.HVALS(J)) GOTO 54
  52     CONTINUE
         HVALS(LOC)=IIA(I)
         LOC=LOC+1
  54     CONTINUE
         DO 56 J=1,LOC-1
            IF (IIB(I).EQ.HVALS(J)) GOTO 58
  56     CONTINUE
         HVALS(LOC)=IIB(I)
         LOC=LOC+1
  58     CONTINUE
  59  CONTINUE
C
      DO 400 LEVEL=1,N-1
         DO 200 I=1,NPLUS1
            ICL=I
            DO 150 ILEV=1,LEVEL
  150       IF (IIB(ILEV).EQ.ICL) ICL=IIA(ILEV)
            NCL = NPLUS1 - LEVEL
            ICLASS(I,NCL-1)=ICL
  200    CONTINUE
  400  CONTINUE
C
      DO 120 I=1,NPLUS1
      DO 120 J=1,N-1
      DO 110 K=2,N
      IF (ICLASS(I,J).NE.HVALS(K)) GOTO 110
         ICLASS(I,J)=K
         GOTO 120
  110 CONTINUE
  120 CONTINUE
C
      RETURN
      END




C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C
C                                                                 C
C  Optimise the variances of a set of groups, by assigning        C
C  the objects in groups such that they are minimally distant     C
C  from group centres.                                            C
C                                                                 C
C  Parameters:                                                    C
C                                                                 C
C  N, M, NG         Numbers of rows, columns, groups,             C
C  A(N,M)           initial data,                                 C
C  MEMGP(N)         group memberships,                            C
C  NGP0             minimum acceptable group cardinality,         C
C  NUMGP(NG)        cardinalities of groups,                      C
C  GPCEN(NG,M)      group centres,                                C
C  COMP(NG)         compactness values for the groups,            C
C  CTOT             sum of these compactnesses,                   C
C  ITMAX            max allowed no. of iterations,                C
C  IERR             error indicator (should be zero).             C
C                                                                 C
C  IERR = 1: invalid group number (<1 or >NG), - is number of     C
C  groups correctly specified?  IERR = 2: a group has < minimum   C
C  allowed number of members, - reduce the number of groups and   C
C  try again.                                                     C
C                                                                 C
C  History                                                        C
C  Author: F. Murtagh, Garching/Munich, Feb. 1986.                C
C  Added parameter ITMAX (formerly hard-wired).  [F.M., Sept. 91] C 
C                                                                 C
C-----------------------------------------------------------------C
        SUBROUTINE MINDST(A,N,M,MEMGP,NGP0,NUMGP,GPCEN,NG,
     X            COMP,CTOT,ITER,ITMAX,IERR)
        DIMENSION       A(N,M), MEMGP(N), NUMGP(NG), GPCEN(NG,M),
     X                  COMP(NG)
C
        BIG = 1.0E+30
        ONE = 0.999
        CMAX = BIG
        ITER = 0
  100   ITER = ITER + 1
        IF (ITER.GT.ITMAX) GOTO 500
        CALL GMEANS(A,N,M,MEMGP,NGP0,NUMGP,GPCEN,NG,IERR)
        CALL COMPCT(A,N,M,NG,MEMGP,GPCEN,COMP,CTOT)
        IF (IERR.NE.0) GOTO 500
        IF (NG.LE.1) GOTO 500
        IF (CTOT.GE.CMAX) GOTO 500
        CMAX = CTOT*ONE
        DO 400 I = 1, N
           X = BIG
           DO 300 K = 1, NG
              Y = 0.0
              DO 200 J = 1, M
                 DIFF = GPCEN(K,J) - A(I,J)
                 Y = Y + DIFF*DIFF
  200         CONTINUE
              IF (Y.GE.X) GOTO 300
              X = Y
              ICL = K
  300      CONTINUE
           MEMGP(I) = ICL
  400   CONTINUE
        GOTO 100
C
  500   RETURN
        END
C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C
C                                                                 C
C  Optimise the variances of a set of groups, by exchanging       C
C  the objects between groups such that they are minimally        C
C  distant from group centres.                                    C
C                                                                 C
C  Parameters:                                                    C
C                                                                 C
C  N, M, NG         Numbers of rows, columns, groups,             C
C  A(N,M)           initial data,                                 C
C  MEMGP(N)         group memberships,                            C
C  NGP0             minimum acceptable group cardinality,         C
C  NUMGP(NG)        cardinalities of groups,                      C
C  GPCEN(NG,M)      group centres,                                C
C  COMP(NG)         compactness values for the groups,            C
C  CTOT             sum of these compactnesses,                   C
C  ITMAX            max allowed no. of iterations,                C
C  IERR             error indicator (should be zero).             C
C                                                                 C
C  IERR = 1: invalid group number (<1 or >NG), - is number of     C
C  groups correctly specified?  IERR = 2: a group has < minimum   C
C  allowed number of members, - reduce the number of groups and   C
C  try again.                                                     C
C                                                                 C
C  History                                                        C
C  Author: F. Murtagh, Garching/Munich, Feb. 1986.                C
C  Added parameter ITMAX (formerly hardwired).  [F.M., Sept. 91]  C
C                                                                 C
C-----------------------------------------------------------------C
        SUBROUTINE EXCH(A,N,M,MEMGP,NGP0,NUMGP,GPCEN,NG,
     X            COMP,CTOT,ITER,ITMAX,IERR)
        DIMENSION       A(N,M), MEMGP(N), NUMGP(NG), GPCEN(NG,M),
     X                  COMP(NG)
C
        BIG = 1.0E+30
        ONE = 0.999
        CALL GMEANS(A,N,M,MEMGP,NGP0,NUMGP,GPCEN,NG,IERR)
        CALL COMPCT(A,N,M,NG,MEMGP,GPCEN,COMP,CTOT)
        IF (IERR.NE.0) GOTO 800
        IF (NG.LE.1) GOTO 800
        ITER = 0
        I = 0
        IS = 0
  100   IS = IS + 1
        IF (IS.GT.N) GOTO 800
  200   I = I + 1
        IF (I.LE.N) GOTO 300
        ITER = ITER + 1
        IF (ITER.GT.ITMAX) GOTO 800
        I = 1
  300   ICL = MEMGP(I)
        NUM = NUMGP(ICL)
        IF (NUM.LE.NGP0) GOTO 100
        V = NUM
        EQ = BIG
        DO 600 K = 1, NG
           X = 0.0
           DO 400 J = 1, M
              DIFF = GPCEN(K,J) - A(I,J)
              X = X + DIFF*DIFF
  400      CONTINUE
           IF (K.NE.ICL) GOTO 500
           FRAC1 = V/(V-1.0)
           EP = X*FRAC1
           GOTO 600
  500      FRAC2 = NUMGP(K)
           FRAC = FRAC2/(FRAC2+1.0)
           EK = FRAC*X
           IF (EK.GE.EQ) GOTO 600
           EQ = EK
           IQ = K
           W = FRAC2
  600   CONTINUE
        IF (EQ.GE.EP*ONE) GOTO 100
        IS = 0
        COMP(ICL) = COMP(ICL) - EP
        COMP(IQ) = COMP(IQ) + EQ
        CTOT = CTOT - EP + EQ
        FRAC1 = 1.0/(V-1.0)
        FRAC2 = 1.0/(W+1.0)
        DO 700 J = 1, M
           VAL = A(I,J)
           GPCEN(ICL,J) = (V*GPCEN(ICL,J)-VAL)*FRAC1
           GPCEN(IQ,J) = (W*GPCEN(IQ,J)+VAL)*FRAC2
  700   CONTINUE
        MEMGP(I) = IQ
        NUMGP(ICL) = NUM - 1
        NUMGP(IQ) = NUMGP(IQ) + 1
        GOTO 200
C
  800   CONTINUE
        RETURN
        END
C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C
C                                                                 C
C  Standardize to zero mean and unit standard deviation.          C
C                                                                 C
C  Parameters:                                                    C
C                                                                 C
C  N, M, NG         Numbers of rows, columns, groups,             C
C  A(N,M)           initial data, replaced by standardized values.C
C                                                                 C
C  Author: F. Murtagh, Garching/Munich, Feb. 1986.                C
C                                                                 C
C-----------------------------------------------------------------C
        SUBROUTINE STND(A,N,M)
        DIMENSION  A(N,M)
C
        DO 500 J = 1, M
           X = 0.0
           DO 100 I = 1, N
              X = X + A(I,J)
  100      CONTINUE
           XBAR = X/FLOAT(N)
           X = 0.0
           DO 200 I = 1, N
              DIFF = A(I,J) - XBAR
              X = X + DIFF*DIFF
  200      CONTINUE
           IF (X.LE.0.0) X = 1.0
           X = 1.0/SQRT(X)
           DO 300 I = 1, N
              A(I,J) = X*(A(I,J)-XBAR)
  300      CONTINUE
  500   CONTINUE
C
        RETURN
        END
C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C
C                                                                 C
C  Determine means of the groups.                                 C
C                                                                 C
C  Parameters:                                                    C
C                                                                 C
C  N, M, NG         Numbers of rows, columns, groups,             C
C  A(N,M)           initial data,                                 C
C  MEMGP(N)         group memberships,                            C
C  NGP0             minimum acceptable group cardinality,         C
C  NUMGP(NG)        cardinalities of groups,                      C
C  GPCEN(NG,M)      group centres,                                C
C  IERR             error indicator (should be zero).             C
C                                                                 C
C  Author: F. Murtagh, Garching/Munich, Feb. 1986.                C
C                                                                 C
C-----------------------------------------------------------------C
        SUBROUTINE GMEANS(A,N,M,MEMGP,NGP0,NUMGP,GPCEN,NG,IERR)
        DIMENSION       A(N,M), MEMGP(N), NUMGP(NG), GPCEN(NG,M)
C
        DO 200 K = 1, NG
           NUMGP(K) = 0
           DO 100 J = 1, M
              GPCEN(K,J) = 0.0
  100      CONTINUE
  200   CONTINUE           
C
        DO 500 I = 1, N
           ICL = MEMGP(I)
           IF (ICL.GE.1.AND.ICL.LE.NG) GOTO 300
              IERR = 1
              RETURN
  300      CONTINUE
           NUMGP(ICL) = NUMGP(ICL) + 1
           DO 400 J = 1, M
              GPCEN(ICL,J) = GPCEN(ICL,J) + A(I,J)
  400      CONTINUE
  500   CONTINUE
C
        DO 800 K = 1, NG
           NUM = NUMGP(K)
           IF (NUM.GE.NGP0) GOTO 600
              IERR = 2
              RETURN
  600      CONTINUE
           X = 1.0/FLOAT(NUM)
           DO 700 J = 1, M
              GPCEN(K,J) = GPCEN(K,J)*X
  700      CONTINUE
  800   CONTINUE
C
        RETURN
        END
C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++C
C                                                                 C
C  Determine compactness of the groups (i.e. their variances).    C
C                                                                 C
C  Parameters:                                                    C
C                                                                 C
C  N, M, NG         Numbers of rows, columns, groups,             C
C  A(N,M)           initial data,                                 C
C  MEMGP(N)         group memberships,                            C
C  GPCEN(NG,M)      group centres,                                C
C  COMP(NG)         variances of groups (output),                 C
C  CTOT             sum of these variances.                       C
C                                                                 C
C  Author: F. Murtagh, Garching/Muenchen, Feb. 1986.              C
C                                                                 C
C-----------------------------------------------------------------C
        SUBROUTINE COMPCT(A,N,M,NG,MEMGP,GPCEN,COMP,CTOT)
        DIMENSION       A(N,M), MEMGP(N), GPCEN(NG,M), COMP(NG)
C
        CTOT = 0.0
        DO 100 K = 1, NG
           COMP(K) = 0.0
  100   CONTINUE
C
        DO 300 I = 1, N
           ICL = MEMGP(I)
           X = 0.0
           DO 200 J = 1, M
              DIFF = GPCEN(ICL,J) - A(I,J)
              X = X + DIFF*DIFF
  200      CONTINUE
           COMP(ICL) = COMP(ICL) + X
           CTOT = CTOT + X
  300   CONTINUE
C
        RETURN
        END
C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C                                  
C  Carry out a PRINCIPAL COMPONENTS ANALYSIS (KARHUNEN-LOEVE EXPANSION).  
C                                           
C  To call: CALL PCA(N,M,DATA,METHOD,A1,W1,W2,W3,W4,A2,IERR) 
C           where
C                                
C                                
C  N, M  : integer dimensions of ...
C  DATA  : input data.              
C          On output, DATA contains in first 7 columns the 
C          projections of the row-points on the first 7 
C          principal components. 
C  METHOD: analysis option.                 
C          = 1: on sums of squares & cross products matrix. 
C               (I.e. no transforming of original data.)
C          = 2: on covariance matrix.    
C               (I.e. original data centered to zero mean.)
C          = 3: on correlation matrix.   
C               (I.e. original data centered, and reduced to unit std.dev.)
C          = 4: on covariance matrix of range-normalized data.
C               (I.e. original data normalized by div. by ranges; then cent'd.)
C          = 5: on Kenkall rank-order correlation matrix.
C          = 6: on Spearman rank-order correlation matrix.
C          = 7: on sample covariance matrix.
C          = 8: on sample correlation matrix.
C  A1    : correlation, covariance or sums of squares & 
C          cross-products matrix, dimensions M * M.            
C          On output, A1 contains in the first 7 columns the 
C          projections of the column-points on the first 7 
C          principal components. 
C  W1,W2 : real vectors of dimension M (see called routines for 
C          use).
C          On output, W1 contains the cumulative percentage 
C          variances associated with the principal components.        
C          Correction Jan. '90: on output, W1 contains the eigenvalues
C          (read with index M-J+1, where running index J is J=1,7)
C  W3,W4 : real vectors of dimension N (used only by Spearman corr. routine)
C  A2    : real array of dimensions M * M (see routines for use).   
C  IERR  : error indicator (normally zero).            
C                                                      
C  Inputs here are N, M, DATA, METHOD (and IERR).
C  Output information is contained in DATA, A1, and W1.  
C  All printed outputs are carried out in easily recognizable sub- 
C  routines called from the first subroutine following.    
C                                                          
C  If IERR > 0, then its value indicates the eigenvalue for which
C  no convergence was obtained.      
C                                 
C  F. Murtagh, ST-ECF/ESA/ESO, Garching/Munich, January 1986. 
C
C  HISTORY
C
C  Initial coding, testing                             [F.M., Jan. 1986]
C  Bug fix - col. projs. overwritten by cum. % var.    [F.M., Jan. 1990]
C  Subr. names altered thoughout, to begin with P and to be <= 6 chars. 
C    long, to avoid conflicts with other similarly named subr. in other
C    prgs. such as corresp. anal. or mult. disc. anal. [F.M., Aug. 1990] 
C  Addition of OPTION=4, and subr. PRANCV              [F.M., May 1991]   
C  Addition of OPTIONs 5 and 6, and assoc. subrs.      [F.M., May 1991]
C  Addition of OPTIONs 7 and 8, and assoc. subrs.      [F.M., Sept. 1991]
C  Remomval of all IPRINT-related subr's. and calls.   [F.M., Feb. 1992]
C                             
C-------------------------------------------------------------------------
        SUBROUTINE PCA(N,M,DATA,METHOD,A,W,FV1,W3,W4,Z,IERR)
        REAL    DATA(N,M), A(M,M), W(M), FV1(M), Z(M,M), W3(N), W4(N)
C
        IF (METHOD.EQ.1) GOTO 100
        IF (METHOD.EQ.2) GOTO 200
        IF (METHOD.EQ.4) GOTO 400
        IF (METHOD.EQ.5) GOTO 500
        IF (METHOD.EQ.6) GOTO 600
        IF (METHOD.EQ.7) GOTO 700
        IF (METHOD.EQ.8) GOTO 800
C       If method.eq.3 or otherwise ...
        GOTO 300
C
C---    Form sums of squares and cross-products matrix.
  100   CONTINUE
        CALL PSCPCL(N,M,DATA,A)
C       Now do the PCA.
        GOTO 1000
C
C---    Form covariance matrix.
  200   CONTINUE
        CALL PCOVCL(N,M,DATA,W,A)
C       Now do the PCA.
        GOTO 1000
C
C---    Construct correlation matrix.
  300   CONTINUE
        CALL PCORCL(N,M,DATA,W,FV1,A)
C       Now do the PCA.
        GOTO 1000
C
C---    Normalize by dividing by ranges; then use covariances.
  400   CONTINUE
        CALL PRANCV(N,M,DATA,W,FV1,A) 
C       Now do the PCA.
        GOTO 1000
C
C---    Construct Kendall rank-order correlation matrix.
  500   CONTINUE
        CALL PKEND(N,M,DATA,A)
C       Now do the PCA.
        GOTO 1000
C
C---    Construct Spearman rank-order correlation matrix.
  600   CONTINUE
        CALL PSPEAR(N,M,DATA,W3,W4,A)
C       Now do the PCA.
        GOTO 1000
C
C---    Form sample covariance matrix.
  700   CONTINUE
        CALL PCOVSA(N,M,DATA,W,A)
C       Now do the PCA.
        GOTO 1000
C
C---    Construct sample correlation matrix.
  800   CONTINUE
        CALL PCORSA(N,M,DATA,W,FV1,A)
C       Now do the PCA.
        GOTO 1000
C
C---    Carry out eigenreduction.
 1000   M2 = M
        CALL PTRED2(M,M2,A,W,FV1,Z)
        CALL PTQL2(M,M2,W,FV1,Z,IERR)
        IF (IERR.NE.0) GOTO 9000
C
C---    Output eigenvalues and eigenvectors.
C
C---    Determine projections and output them.
        CALL PPROJX(N,M,DATA,Z,FV1)
        CALL PPROJY(M,W,A,Z,FV1)
C
 9000   RETURN  
        END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C                           
C  Determine correlations of columns.
C  First determine the means of columns, storing in WORK1.
C                                         
C--------------------------------------------------------
        SUBROUTINE PCORCL(N,M,DATA,WORK1,WORK2,OUT)
        DIMENSION       DATA(N,M), OUT(M,M), WORK1(M), WORK2(M)
        DATA            EPS/1.E-10/
C
        DO 30 J = 1, M
           WORK1(J) = 0.0
           DO 20 I = 1, N
              WORK1(J) = WORK1(J) + DATA(I,J)
   20      CONTINUE
           WORK1(J) = WORK1(J)/FLOAT(N)
   30   CONTINUE
C
C          Next det. the std. devns. of cols., storing in WORK2.
C
        DO 50 J = 1, M
           WORK2(J) = 0.0
           DO 40 I = 1, N
              WORK2(J) = WORK2(J) + (DATA(I,J)
     X                   -WORK1(J))*(DATA(I,J)-WORK1(J))
   40      CONTINUE
           WORK2(J) = WORK2(J)/FLOAT(N)
           WORK2(J) = SQRT(WORK2(J))
           IF (WORK2(J).LE.EPS) WORK2(J) = 1.0
   50   CONTINUE
C
C          Now centre and reduce the column points.
C
        DO 70 I = 1, N
           DO 60 J = 1, M
              DATA(I,J) = (DATA(I,J)
     X                    -WORK1(J))/(SQRT(FLOAT(N))*WORK2(J))
   60      CONTINUE
   70   CONTINUE
C
C          Finally calc. the cross product of the data matrix.
C
        DO 100 J1 = 1, M-1
           OUT(J1,J1) = 1.0
           DO 90 J2 = J1+1, M
              OUT(J1,J2) = 0.0
              DO 80 I = 1, N
                 OUT(J1,J2) = OUT(J1,J2) + DATA(I,J1)*DATA(I,J2)
   80         CONTINUE
              OUT(J2,J1) = OUT(J1,J2)
   90      CONTINUE
  100   CONTINUE
        OUT(M,M) = 1.0
C
        RETURN
        END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++
C                       
C  Determine covariances of columns. 
C  First determine the means of columns, storing in WORK.
C                                
C------------------------------------------------------
        SUBROUTINE PCOVCL(N,M,DATA,WORK,OUT)
        DIMENSION       DATA(N,M), OUT(M,M), WORK(M)
C
        DO 30 J = 1, M
           WORK(J) = 0.0
           DO 20 I = 1, N
              WORK(J) = WORK(J) + DATA(I,J)
   20      CONTINUE
           WORK(J) = WORK(J)/FLOAT(N)
   30   CONTINUE
C
C          Now centre the column points.
C
        DO 50 I = 1, N
           DO 40 J = 1, M
              DATA(I,J) = DATA(I,J)-WORK(J)
   40      CONTINUE
   50   CONTINUE
C
C          Finally calculate the cross product matrix of the 
C          redefined data matrix.
C
        DO 80 J1 = 1, M
           DO 70 J2 = J1, M
              OUT(J1,J2) = 0.0
              DO 60 I = 1, N
                 OUT(J1,J2) = OUT(J1,J2) + DATA(I,J1)*DATA(I,J2)
   60         CONTINUE
              OUT(J2,J1) = OUT(J1,J2)
   70      CONTINUE
   80   CONTINUE
C
        RETURN
        END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C                           
C  Determine sample correlations of columns.
C  First determine the means of columns, storing in WORK1.
C                                         
C--------------------------------------------------------
        SUBROUTINE PCORSA(N,M,DATA,WORK1,WORK2,OUT)
        DIMENSION       DATA(N,M), OUT(M,M), WORK1(M), WORK2(M)
        DATA            EPS/1.E-10/
C
        DO 30 J = 1, M
           WORK1(J) = 0.0
           DO 20 I = 1, N
              WORK1(J) = WORK1(J) + DATA(I,J)
   20      CONTINUE
           WORK1(J) = WORK1(J)/FLOAT(N-1)
   30   CONTINUE
C
C          Next det. the std. devns. of cols., storing in WORK2.
C
        DO 50 J = 1, M
           WORK2(J) = 0.0
           DO 40 I = 1, N
              WORK2(J) = WORK2(J) + (DATA(I,J)
     X                   -WORK1(J))*(DATA(I,J)-WORK1(J))
   40      CONTINUE
           WORK2(J) = WORK2(J)/FLOAT(N-1)
           WORK2(J) = SQRT(WORK2(J))
           IF (WORK2(J).LE.EPS) WORK2(J) = 1.0
   50   CONTINUE
C
C          Now centre and reduce the column points.
C
        DO 70 I = 1, N
           DO 60 J = 1, M
              DATA(I,J) = (DATA(I,J)
     X                    -WORK1(J))/(SQRT(FLOAT(N-1))*WORK2(J))
   60      CONTINUE
   70   CONTINUE
C
C          Finally calc. the cross product of the data matrix.
C
        DO 100 J1 = 1, M-1
           OUT(J1,J1) = 1.0
           DO 90 J2 = J1+1, M
              OUT(J1,J2) = 0.0
              DO 80 I = 1, N
                 OUT(J1,J2) = OUT(J1,J2) + DATA(I,J1)*DATA(I,J2)
   80         CONTINUE
              OUT(J2,J1) = OUT(J1,J2)
   90      CONTINUE
  100   CONTINUE
        OUT(M,M) = 1.0
C
        RETURN
        END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++
C                       
C  Determine sample covariances of columns. 
C  First determine the means of columns, storing in WORK.
C                                
C------------------------------------------------------
        SUBROUTINE PCOVSA(N,M,DATA,WORK,OUT)
        DIMENSION       DATA(N,M), OUT(M,M), WORK(M)
C
        DO 30 J = 1, M
           WORK(J) = 0.0
           DO 20 I = 1, N
              WORK(J) = WORK(J) + DATA(I,J)
   20      CONTINUE
           WORK(J) = WORK(J)/FLOAT(N-1)
   30   CONTINUE
C
C          Now centre the column points.
C
        DO 50 I = 1, N
           DO 40 J = 1, M
              DATA(I,J) = DATA(I,J)-WORK(J)
   40      CONTINUE
   50   CONTINUE
C
C          Finally calculate the cross product matrix of the 
C          redefined data matrix.
C
        DO 80 J1 = 1, M
           DO 70 J2 = J1, M
              OUT(J1,J2) = 0.0
              DO 60 I = 1, N
                 OUT(J1,J2) = OUT(J1,J2) + DATA(I,J1)*DATA(I,J2)
   60         CONTINUE
              OUT(J2,J1) = OUT(J1,J2)
   70      CONTINUE
   80   CONTINUE
C
        RETURN
        END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++
C                       
C  Determine covariances of columns, having first 
C  normalized by dividing by ranges. 
C  First determine max and min of columns, storing in 
C  WORK1 and WORK2.
C                                
C------------------------------------------------------
        SUBROUTINE PRANCV(N,M,DATA,WORK1,WORK2,OUT)
        DIMENSION       DATA(N,M), OUT(M,M), WORK1(M), WORK2(M)
C
        DO 30 J = 1, M
           WORK1(J) = -10000.0
           WORK2(J) =  10000.0
           DO 20 I = 1, N
              IF (DATA(I,J).GT.WORK1(J)) WORK1(J) = DATA(I,J)
              IF (DATA(I,J).LT.WORK2(J)) WORK2(J) = DATA(I,J)
   20      CONTINUE
C          Range:
           WORK1(J) = WORK1(J)-WORK2(J)
   30   CONTINUE
C
C          Now normalize the column points.
C
        DO 50 J = 1, M
           WORK2(J) = 0.0
           DO 40 I = 1, N
              DATA(I,J) = DATA(I,J)/WORK1(J)
              WORK2(J) = WORK2(J) + DATA(I,J)
   40      CONTINUE
           WORK2(J) = WORK2(J)/FLOAT(N)
   50   CONTINUE
C
        DO 58 I = 1, N
           DO 56 J = 1, M
              DATA(I,J) = DATA(I,J)-WORK2(J)
 56        CONTINUE
 58     CONTINUE
C
C          Finally calculate the cross product matrix of the 
C          redefined data matrix.
C
        DO 80 J1 = 1, M
           DO 70 J2 = J1, M
              OUT(J1,J2) = 0.0
              DO 60 I = 1, N
                 OUT(J1,J2) = OUT(J1,J2) + DATA(I,J1)*DATA(I,J2)
   60         CONTINUE
              OUT(J2,J1) = OUT(J1,J2)
   70      CONTINUE
   80   CONTINUE
C
        RETURN
        END
C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C                                                         
C  Determine sums of squares and cross-products of columns.
C                                            
C---------------------------------------------------------
        SUBROUTINE PSCPCL(N,M,DATA,OUT)
        DIMENSION       DATA(N,M), OUT(M,M)
C
        DO 30 J1 = 1, M
           DO 20 J2 = J1, M
              OUT(J1,J2) = 0.0
              DO 10 I = 1, N
                 OUT(J1,J2) = OUT(J1,J2) + DATA(I,J1)*DATA(I,J2)
   10         CONTINUE
              OUT(J2,J1) = OUT(J1,J2)
   20      CONTINUE
   30   CONTINUE
C
        RETURN
        END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C                                   
C Reduce a real, symmetric matrix to a symmetric, tridiagonal
C matrix.                         
C                                 
C To call:    CALL PTRED2(NM,N,A,D,E,Z)    where
C                                     
C NM = row dimension of A and Z;      
C N = order of matrix A (will always be <= NM);
C A = symmetric matrix of order N to be reduced to tridiag. form;
C D = vector of dim. N containing, on output, diagonal elts. of
C     tridiagonal matrix.               
C E = working vector of dim. at least N-1 to contain subdiagonal
C     elements.                          
C Z = matrix of dims. NM by N containing, on output, orthogonal
C     transformation matrix producing the reduction. 
C                                            
C Normally a call to PTQL2 will follow the call to PTRED2 in order to
C produce all eigenvectors and eigenvalues of matrix A.
C                                        
C Algorithm used: Martin et al., Num. Math. 11, 181-195, 1968. 
C                                 
C Reference: Smith et al., Matrix Eigensystem Routines - EISPACK
C Guide, Lecture Notes in Computer Science 6, Springer-Verlag, 
C 1976, pp. 489-494.                     
C                                        
C----------------------------------------------------------------
        SUBROUTINE PTRED2(NM,N,A,D,E,Z)
        REAL A(NM,N),D(N),E(N),Z(NM,N)
C
        DO 100 I = 1, N
           DO 100 J = 1, I
              Z(I,J) = A(I,J)
  100   CONTINUE
        IF (N.EQ.1) GOTO 320
        DO 300 II = 2, N
           I = N + 2 - II
           L = I - 1
           H = 0.0
           SCALE = 0.0
           IF (L.LT.2) GOTO 130
           DO 120 K = 1, L
              SCALE = SCALE + ABS(Z(I,K))
  120      CONTINUE
           IF (SCALE.NE.0.0) GOTO 140
  130      E(I) = Z(I,L)
           GOTO 290
  140      DO 150 K = 1, L
              Z(I,K) = Z(I,K)/SCALE
              H = H + Z(I,K)*Z(I,K)
  150      CONTINUE
C
           F = Z(I,L)
           G = -SIGN(SQRT(H),F)
           E(I) = SCALE * G
           H = H - F * G
           Z(I,L) = F - G
           F = 0.0
C
           DO 240 J = 1, L
              Z(J,I) = Z(I,J)/H
              G = 0.0
C             Form element of A*U.
              DO 180 K = 1, J
                 G = G + Z(J,K)*Z(I,K)
  180         CONTINUE
              JP1 = J + 1
              IF (L.LT.JP1) GOTO 220
              DO 200 K = JP1, L
                 G = G + Z(K,J)*Z(I,K)
  200         CONTINUE
C             Form element of P where P = I - U U' / H .
  220         E(J) = G/H
              F = F + E(J) * Z(I,J)
  240      CONTINUE
           HH = F/(H + H)
C          Form reduced A.
           DO 260 J = 1, L
              F = Z(I,J)
              G = E(J) - HH * F
              E(J) = G
              DO 250 K = 1, J
                 Z(J,K) = Z(J,K) - F*E(K) - G*Z(I,K)
  250         CONTINUE
  260      CONTINUE
  290      D(I) = H
  300   CONTINUE
  320   D(1) = 0.0
        E(1) = 0.0
C       Accumulation of transformation matrices.
        DO 500 I = 1, N
           L = I - 1
           IF (D(I).EQ.0.0) GOTO 380
           DO 360 J = 1, L
              G = 0.0
              DO 340 K = 1, L
                 G = G + Z(I,K) * Z(K,J)
  340         CONTINUE
              DO 350 K = 1, L
                 Z(K,J) = Z(K,J) - G * Z(K,I)
  350         CONTINUE
  360      CONTINUE
  380      D(I) = Z(I,I)
           Z(I,I) = 1.0
           IF (L.LT.1) GOTO 500
           DO 400 J = 1, L
              Z(I,J) = 0.0
              Z(J,I) = 0.0
  400      CONTINUE
  500   CONTINUE
C
        RETURN
        END
C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C              
C Determine eigenvalues and eigenvectors of a symmetric,
C tridiagonal matrix.     
C                         
C To call:    CALL PTQL2(NM,N,D,E,Z,IERR)    where
C                             
C NM = row dimension of Z;    
C N = order of matrix Z;      
C D = vector of dim. N containing, on output, eigenvalues;
C E = working vector of dim. at least N-1;      
C Z = matrix of dims. NM by N containing, on output, eigenvectors;
C IERR = error, normally 0, but 1 if no convergence.    
C                      
C Normally the call to PTQL2 will be preceded by a call to PTRED2 in 
C order to set up the tridiagonal matrix.   
C                     
C Algorithm used: QL method of Bowdler et al., Num. Math. 11,
C 293-306, 1968.                   
C                                 
C Reference: Smith et al., Matrix Eigensystem Routines - EISPACK 
C Guide, Lecture Notes in Computer Science 6, Springer-Verlag,
C 1976, pp. 468-474.                 
C                                    
C--------------------------------------------------------------
        SUBROUTINE PTQL2(NM,N,D,E,Z,IERR)
        REAL    D(N), E(N), Z(NM,N)
        DATA    EPS/1.E-12/
C
        IERR = 0
        IF (N.EQ.1) GOTO 1001
        DO 100 I = 2, N
           E(I-1) = E(I)
  100   CONTINUE
        F = 0.0
        B = 0.0
        E(N) = 0.0
C
        DO 240 L = 1, N
           J = 0
           H = EPS * (ABS(D(L)) + ABS(E(L)))
           IF (B.LT.H) B = H
C          Look for small sub-diagonal element.
           DO 110 M = L, N
              IF (ABS(E(M)).LE.B) GOTO 120
C             E(N) is always 0, so there is no exit through
C             the bottom of the loop.
  110      CONTINUE
  120      IF (M.EQ.L) GOTO 220
  130      IF (J.EQ.30) GOTO 1000
           J = J + 1
C          Form shift.
           L1 = L + 1
           G = D(L)
           P = (D(L1)-G)/(2.0*E(L))
           R = SQRT(P*P+1.0)
           D(L) = E(L)/(P+SIGN(R,P))
           H = G-D(L)
C
           DO 140 I = L1, N
              D(I) = D(I) - H
  140      CONTINUE
C
           F = F + H
C          QL transformation.
           P = D(M)
           C = 1.0
           S = 0.0
           MML = M - L
C
           DO 200 II = 1, MML
              I = M - II
              G = C * E(I)
              H = C * P
              IF (ABS(P).LT.ABS(E(I))) GOTO 150
              C = E(I)/P
              R = SQRT(C*C+1.0)
              E(I+1) = S * P * R
              S = C/R
              C = 1.0/R
              GOTO 160
  150         C = P/E(I)
              R = SQRT(C*C+1.0)
              E(I+1) = S * E(I) * R
              S = 1.0/R
              C = C * S
  160         P = C * D(I) - S * G
              D(I+1) = H + S * (C * G + S * D(I))
C             Form vector.
              DO 180 K = 1, N
                 H = Z(K,I+1)
                 Z(K,I+1) = S * Z(K,I) + C * H
                 Z(K,I) = C * Z(K,I) - S * H
  180         CONTINUE
  200      CONTINUE
           E(L) = S * P
           D(L) = C * P
           IF (ABS(E(L)).GT.B) GOTO 130
  220      D(L) = D(L) + F
  240   CONTINUE
C
C       Order eigenvectors and eigenvalues.
        DO 300 II = 2, N
           I = II - 1
           K = I
           P = D(I)
           DO 260 J = II, N
              IF (D(J).GE.P) GOTO 260
              K = J
              P = D(J)
  260      CONTINUE
           IF (K.EQ.I) GOTO 300
           D(K) = D(I)
           D(I) = P
           DO 280 J = 1, N
              P = Z(J,I)
              Z(J,I) = Z(J,K)
              Z(J,K) = P
  280      CONTINUE
  300   CONTINUE
C
        GOTO 1001
C       Set error - no convergence after 30 iterns.
 1000   IERR = L
 1001   RETURN
        END  
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C              
C  Form projections of row-points on first 7 principal components.
C                     
C--------------------------------------------------------------
        SUBROUTINE PPROJX(N,M,DATA,EVEC,VEC)
        REAL    DATA(N,M), EVEC(M,M), VEC(M)
C
        NUM = MIN0(M,7)
        DO 300 K = 1, N
           DO 50 L = 1, M
              VEC(L) = DATA(K,L)
   50      CONTINUE
           DO 200 I = 1, NUM
              DATA(K,I) = 0.0
              DO 100 J = 1, M
                 DATA(K,I) = DATA(K,I) + VEC(J) *
     X                                   EVEC(J,M-I+1)
  100         CONTINUE
  200      CONTINUE
  300   CONTINUE
C
        RETURN
        END
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C                   
C  Determine projections of column-points on 7 prin. components.
C                           
C--------------------------------------------------------------
        SUBROUTINE PPROJY(M,EVALS,A,Z,VEC)
        REAL    EVALS(M), A(M,M), Z(M,M), VEC(M)
C
        NUM = MIN0(M,7)
        DO 300 J1 = 1, M
           DO 50 L = 1, M
              VEC(L) = A(J1,L)
   50      CONTINUE
           DO 200 J2 = 1, NUM
              A(J1,J2) = 0.0
              DO 100 J3 = 1, M
                 A(J1,J2) = A(J1,J2) + VEC(J3)*Z(J3,M-J2+1)
  100         CONTINUE
              IF (EVALS(M-J2+1).GT.0.00005) A(J1,J2) = 
     X                       A(J1,J2)/SQRT(EVALS(M-J2+1))
              IF (EVALS(M-J2+1).LE.0.00005) A(J1,J2) = 0.0 
  200      CONTINUE
  300   CONTINUE
C
        RETURN
        END
C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C  
C   Determine Spearman (rank-order) correlations.
C   Adapted from: Numerical Recipes, Press et al.
C
C-----------------------------------------------------------
        SUBROUTINE PSPEAR(N,M,DATA,WKSP1,WKSP2,A)
        DIMENSION DATA(N,M), WKSP1(N), WKSP2(N), A(M,M)
        DO 900 J1 = 1, M-1
           A(J1,J1) = 1.0
           DO 800 J2 = J1+1, M
              DO 100 I=1,N
                 WKSP1(I)=DATA(I,J1)
                 WKSP2(I)=DATA(I,J2)
 100          CONTINUE
              CALL PSORT(N,WKSP1,WKSP2)
              CALL PRANK(N,WKSP1,SF)
              CALL PSORT(N,WKSP2,WKSP1)
              CALL PRANK(N,WKSP2,SG)
              D=0.
              DO 200 I=1,N
                 D=D+(WKSP1(I)-WKSP2(I))**2
 200          CONTINUE
              EN=N
              EN3N=EN**3-EN
C                AVED=EN3N/6.-(SF+SG)/12.
              FAC=(1.-SF/EN3N)*(1.-SG/EN3N)
C                VARD=((EN-1.)*EN**2*(EN+1.)**2/36.)*FAC
C                ZD=(D-AVED)/SQRT(VARD)
C                PROBD=ERFCC(ABS(ZD)/1.4142136)
              RS=(1.-(6./EN3N)*(D+0.5*(SF+SG)))/FAC
C                T=RS*SQRT((EN-2.)/((1.+RS)*(1.-RS)))
C                DF=EN-2.
C                PROBRS=BETAI(0.5*DF,0.5,DF/(DF+T**2))
              A(J1,J2) = RS
              A(J2,J1) = RS
 800       CONTINUE
 900    CONTINUE
        RETURN
        END
C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        SUBROUTINE PSORT(N,RA,RB)
        DIMENSION RA(N),RB(N)
        L=N/2+1
        IR=N
 10     CONTINUE
              IF (L.GT.1) THEN
                        L=L-1
                        RRA=RA(L)
                        RRB=RB(L)
               ELSE
                        RRA=RA(IR)
                        RRB=RB(IR)
                        RA(IR)=RA(1)
                        RB(IR)=RB(1)
                        IR=IR-1
                        IF (IR.EQ.1) THEN
                                RA(1)=RRA
                                RB(1)=RRB
                                RETURN
                        ENDIF
                ENDIF
                I=L
                J=L+L
 20             IF (J.LE.IR) THEN
                        IF (J.LT.IR) THEN
                           IF (RA(J).LT.RA(J+1)) J=J+1
                        ENDIF
                        IF(RRA.LT.RA(J))THEN
                                RA(I)=RA(J)
                                RB(I)=RB(J)
                                I=J
                                J=J+J
                        ELSE
                                J=IR+1
                        ENDIF
                GO TO 20
                ENDIF
                RA(I)=RRA
                RB(I)=RRB
        GO TO 10
        END
C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        SUBROUTINE PRANK(N,W,S)
        DIMENSION W(N)
        S=0.
        J=1
 1      IF (J.LT.N) THEN
           IF (W(J+1).NE.W(J)) THEN
                        W(J)=J
                        J=J+1
            ELSE
                        DO 100 JT=J+1,N
                               IF (W(JT).NE.W(J)) GO TO 2
 100                    CONTINUE
                        JT=N+1
 2                      RANK=0.5*(J+JT-1)
                        DO 200 JI=J,JT-1
                               W(JI)=RANK
 200                    CONTINUE
                        T=JT-J
                        S=S+T**3-T
                        J=JT
             ENDIF
        GO TO 1
        ENDIF
        IF(J.EQ.N)W(N)=N
        RETURN
        END
C+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C
C Kendall (rank-order) correlations
C
C-----------------------------------------------------------
        SUBROUTINE PKEND(N,M,DATA,A)
        DIMENSION DATA(N,M), A(M,M)
        DO 900 J1 = 1, M-1
           A(J1,J1) = 1.0
           DO 800 J2 = J1+1, M
              N1=0
              N2=0
              IS=0
               DO 600 J=1,N-1
                  DO 500 K=J+1,N
                        A1=DATA(J,J1)-DATA(K,J1)
                        A2=DATA(J,J2)-DATA(K,J2)
                        AA=A1*A2
                        IF(AA.NE.0.)THEN
                                N1=N1+1
                                N2=N2+1
                                IF(AA.GT.0.)THEN
                                        IS=IS+1
                                ELSE
                                        IS=IS-1
                                ENDIF
                        ELSE
                                IF(A1.NE.0.)N1=N1+1
                                IF(A2.NE.0.)N2=N2+1
                        ENDIF
 500            CONTINUE
 600        CONTINUE
            TAU=FLOAT(IS)/SQRT(FLOAT(N1)*FLOAT(N2))
C           VAR=(4.*N+10.)/(9.*N*(N-1.))
C           Z=TAU/SQRT(VAR)
C           PROB=ERFCC(ABS(Z)/1.4142136)
            A(J1,J2) = TAU
            A(J2,J1) = TAU
 800      CONTINUE
 900    CONTINUE
        RETURN
        END
      subroutine sammon(n,m,p,x,y,ndis,dstar,d,
     .                  alpha,maxit,diag,iter,tol,err)

c     SAMMON MAPPING (NON-METRIC MULTIDIMENSIONAL SCALING). 
c     Author: F. Murtagh, May 1992.

      integer   n,m,p,i,j,k,iter,maxit,diag
      real      alpha,tol,err
      dimension x(n,m), y(n,p), dstar(ndis), d(ndis)
c
      iter = 0
c
      c = 0.0
      do i = 1, n-1
         do j = i+1, n
            ind = n*(i-1) - i*(i-1)/2 + j - i
            c = c + dstar(ind)
         enddo
      enddo
c
 100  continue
      iter = iter + 1

      do i = 1, n
         do k = 1, p
c
            dedy   = 0.0
            d2edy2 = 0.0
            do j   = 1, n
c
               if (j.eq.i) goto 200
               if (i.lt.j) then
                           ind = n*(i-1) - i*(i-1)/2 + j - i
               else
                           ind = n*(j-1) - j*(j-1)/2 + i - j
               endif
               if (dstar(ind).le.0.001) goto 200
               if (d(ind).le.0.001) goto 200
c
               xnumer1 = y(i,k)-y(j,k)
               xnumsq  = xnumer1*xnumer1
               xnumer2 = dstar(ind)-d(ind)
               denom1 = dstar(ind)*d(ind)
c
               dedy = dedy + xnumer1*xnumer2/denom1
c
               d2edy2 = d2edy2 + (1.0/denom1)*
     .            (xnumer2-(xnumsq/d(ind))*(1.0+xnumer2/d(ind)))
c
 200           continue
            enddo
            dedy = -dedy
            d2edy2 = abs(-d2edy2)
            deltay = dedy/d2edy2
c
            y(i,k) = y(i,k) - alpha*deltay
         enddo
       enddo
c
       do i = 1, n-1
          do j = i+1, n
            ind = n*(i-1) - i*(i-1)/2 + j - i
             d(ind) = 0.0
             do k = 1, p
                d(ind) = d(ind) + (y(i,k)-y(j,k))*(y(i,k)-y(j,k))
             enddo
             d(ind) = sqrt(d(ind))
          enddo
       enddo
c
       err = 0.0
       do i = 1, n-1
          do j = i+1, n
            ind = n*(i-1) - i*(i-1)/2 + j - i
             if (dstar(ind).gt.0.001)
     .         err = err + 
     .         ((dstar(ind)-d(ind))*(dstar(ind)-d(ind)))/dstar(ind)
          enddo
       enddo
       err = err/c
       if (diag.eq.1) call realpr(' ',1,err,1)
c
c       write(6,*) ' Error: ',err
       if (err.le.tol) return
       if (iter.ge.maxit) return
       goto 100
c
       end



